# SPDX-License-Identifier: LGPL-2.1-only OR BSD-3-Clause */
##########################################################################
#
# This file is provided under a dual license.  When you use or
# distribute this software, you may choose to be licensed under
# version 2.1 of the GNU Lesser General Public License
# ("LGPLv2.1 License") or BSD License.
#
# LGPLv2.1 License
#
# Copyright(C) 2019 MediaTek Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of version 2.1 of the GNU Lesser General Public
# License as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See https://www.gnu.org/licenses/old-licenses/lgpl-2.1.html for more
# details.
#
# BSD LICENSE
#
# Copyright(C) 2019 MediaTek Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
###########################################################################

ifndef TARGET
export TARGET = 5891
endif

ifndef DRIVER_DIR
export DRIVER_DIR := $(subst mt,,$(TARGET_IC))_driver
endif

ifndef DRV_ROOT
DRV_ROOT = .
endif

always += $(patsubst %.c,%.o,$(SRC))
ifeq "$(SUSPEND_TO_DRAM)" "true"
always += $(patsubst %.S,%.o,$(ASM_SRC))
endif
obj-m += $(addsuffix /,$(SUB_COMPONENTS))
obj-m += $(addsuffix /,$(BRDCST_IF_COMPONENTS))
obj-m += $(addsuffix /,$(OPTIONAL_SUB_COMPONENTS))

ifdef CONFIG_FUNCTION_TRACER
CFLAGS_REMOVE_core1.o = -pg
CFLAGS_core1.o = -fomit-frame-pointer
endif

include $(DRV_ROOT)/config.mak
include $(DRV_ROOT)/target.mak
include $(DRV_ROOT)/nptv.mak

include $(DRV_ROOT)/drv_opt/$(COMPANY)/$(MODEL).def

CHK_FLAG = $(shell if [ -e $(src)/NONE_CHECK_WARN -o -h $(src)/NONE_CHECK_WARN ]; then echo "false"; else echo "true"; fi)
CHK_OPT_LVL_FLAG = $(shell if [ -e $(src)/OPTIMIZE_LVL -o -h $(src)/OPTIMIZE_LVL ]; then echo "true"; else echo "false"; fi)

ifneq "$(CHK_OPT_LVL_FLAG)" "true"
EXTRA_CFLAGS_R = $(subst -O0,,$(DEFINES) $(CC_INC) $(DRIVER_INC))
#previous version, -O0 is changed to -O1; In this version, I remove -O0. Code size will be small even if OPTIMIZE_LVL=-O0
else
EXTRA_CFLAGS_R = $(DEFINES) $(CC_INC) $(DRIVER_INC)
endif
ifeq "$(CHK_FLAG)" "false"
EXTRA_CFLAGS = $(subst -Werror,-Wall,$(EXTRA_CFLAGS_R))
else
EXTRA_CFLAGS = $(EXTRA_CFLAGS_R)
endif

ifeq "$(TOOL_CHAIN)" "4.8.2"
EXTRA_CFLAGS += -Wno-address -Wno-maybe-uninitialized -Wno-unused-local-typedefs
endif

EXTRA_AFLAGS += $(ASM_DEFINES)

ifeq "$(RLS_CUSTOM_BUILD)" "true"
ifndef VM_LINUX_ROOT
	VM_LINUX_ROOT := $(word 1, $(subst /vm_linux/,/vm_linux /, $(shell pwd)))
endif

ifndef LIB_CONTENT
ifdef BUILD_MOUDLE_TYPE
	LIB_CONTENT= $(BUILD_MOUDLE_TYPE).content
else
	LIB_CONTENT= lib.content
endif
endif

ifeq "$(BUILD_CFG)" "debug"
OBJECT_TYPE = debug
else
ifeq "$(BUILD_CFG)" "cli"
OBJECT_TYPE = cli
else
OBJECT_TYPE = rel
endif
endif


TARGET_DRIVER := $(subst mt,,$(TARGET_IC))_driver
ifndef RLS_OBJ_ROOT
RLS_OBJ_ROOT = $(VM_LINUX_ROOT)/mtk_obj/$(CUSTOMER)/$(MODEL_NAME)/$(OBJECT_TYPE)/driver
endif
SRC_ROOT = $(DTV_ROOT_OUTPUT)/$(CUSTOMER)/$(MODEL_NAME)/$(OBJECT_TYPE)/obj/driver/linux_core/driver/ko
ifdef COMP_PATH_EXT
ifeq "$(findstring $(COMP_PATH_EXT),$(RLS_OBJ_ROOT))" ""
RLS_OBJ_ROOT := $(RLS_OBJ_ROOT)/$(TARGET_IC)/$(TARGET_DRIVER)/brdcst_if/$(COMP_PATH_EXT)
SRC_ROOT := $(SRC_ROOT)/$(TARGET_IC)/$(TARGET_DRIVER)/brdcst_if
endif
endif
$(SRC_ROOT)/%.o : $(RLS_OBJ_ROOT)/%.o
	@if [ ! -e $(LIB_CONTENT) ] || [ "`grep -c -w $< $(LIB_CONTENT)`" == '0' ]; then \
		echo '  MTK_OBJ $<'; \
		echo $< >> $(LIB_CONTENT); \
	fi
endif


