/* SPDX-License-Identifier: LGPL-2.1-only OR BSD-3-Clause */
/******************************************************************************
 *
 * This file is provided under a dual license.  When you use or
 * distribute this software, you may choose to be licensed under
 * version 2.1 of the GNU Lesser General Public License
 * ("LGPLv2.1 License") or BSD License.
 *
 * LGPLv2.1 License
 *
 * Copyright(C) 2019 MediaTek Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2.1 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See https://www.gnu.org/licenses/old-licenses/lgpl-2.1.html for more
 * details.
 *
 * BSD LICENSE
 *
 * Copyright(C) 2019 MediaTek Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *  * Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *****************************************************************************/

/*-----------------------------------------------------------------------------
 *
 * $Author: brianpc.huang $
 * $Date: 2015/05/29 $
 * $RCSfile: mtal_sbtl_if.h,v $
 * $Revision: #1 $
 *
 *---------------------------------------------------------------------------*/

/** @file mtal_sbtl_if.h
 *  MTAL external subtitle interface header file.
 */

#include "x_os.h"
#include "u_handle.h"

#include "x_common.h"
#include "x_mm_common.h"
#include "x_mm_sbtl_engine.h"

//#include "../../../target/mt5391/5391_driver/fm/u_fm.h"
#define FMR_OK     0
#define FMR_EOF    (-1)

#define EXT_SBTL_STR_BUF_SIZE  512  // this must exactly match that in mtmplayer_lib.c

#define MST_MASK_ENABLE  0x00000001
#define MST_MASK_TRK_NO  0x00000002
#define MST_MASK_SYNC    0x00000004


typedef enum  // copied from u_mm_sbtl_engine.h, tmp solution for fixing build error
{
    MM_SBTL_TYPE_DIVX,
    MM_SBTL_TYPE_VOBSUB,
    MM_SBTL_TYPE_TEXT,
    MM_SBTL_TYPE_TEXT_SRT,
    MM_SBTL_TYPE_TEXT_SAMI,
    MM_SBTL_TYPE_TEXT_SVSUB,
    MM_SBTL_TYPE_TEXT_MPLAYER2,
    MM_SBTL_TYPE_TEXT_SSA,
    MM_SBTL_TYPE_TEXT_MDVDSUB,
    MM_SBTL_TYPE_TEXT_TXT,
    MM_SBTL_TYPE_TEXT_POWERDIVX,
    MM_SBTL_TYPE_TEXT_DVDSS,
    MM_SBTL_TYPE_END
} MM_SBTL_TYPE_T;


typedef struct
{
    BOOL           fgExist;
    BOOL           fgEnable;
    BOOL           fgActive;
    BOOL           fgToChgTrack;
    MM_SBTL_TYPE_T eExtSbtlType;
    UINT32         u4ExtSbtlSize;
    UINT16         u2TrackNo;
    UINT16         u2TrackNs;
    INT8           i1Sync;
} EXT_SBTL_SETTING_T;


BOOL fgMtalSbtlTextOpen(void);

BOOL fgMtalSbtlTextClose(void);

BOOL fgMtalSbtlTextStart(void);

BOOL fgMtalSbtlTextStop(void);

BOOL fgMtalSbtlTextChgTrack(void);

void vMtalSbtlTextBufInit(UINT8 *pu1SubtitleBuffer);

void vMtalSbtlTextAssign(BOOL fgExist, MM_SBTL_TYPE_T eType, UINT32 u4Size);

void vMtalSbtlTextSet(UINT32 u4Mask, BOOL fgEnable, UINT16 u2TrackNo, INT8 i1Sync);

BOOL fgIsMtalSbtlExisted(void);

BOOL fgIsMtalSbtlEnabled(void);

UINT16 u2MtalSbtlGetTrackNs(void);

void vMtalSbtlTextCheck(void);

INT32 i4MtalSbtlGetSync(void);

