/* SPDX-License-Identifier: LGPL-2.1-only OR BSD-3-Clause */
/******************************************************************************
 *
 * This file is provided under a dual license.  When you use or
 * distribute this software, you may choose to be licensed under
 * version 2.1 of the GNU Lesser General Public License
 * ("LGPLv2.1 License") or BSD License.
 *
 * LGPLv2.1 License
 *
 * Copyright(C) 2019 MediaTek Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2.1 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See https://www.gnu.org/licenses/old-licenses/lgpl-2.1.html for more
 * details.
 *
 * BSD LICENSE
 *
 * Copyright(C) 2019 MediaTek Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *  * Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *****************************************************************************/

/*-----------------------------------------------------------------------------
 * $RCSfile: c_time_msrt.c,v $
 * $Revision: #1 $
 * $Date: 2015/05/29 $
 * $Author: brianpc.huang $
 * $CCRevision: /main/DTV_X_ATSC/2 $
 * $SWAuthor: Iolo Tsai $
 * $MD5HEX: 7995b837b0cb90a748f8f663c8ccbce5 $
 *
 * Description: 
 *         This header file contains specific implementation.
 *---------------------------------------------------------------------------*/


/*-----------------------------------------------------------------------------
                    include files
 ----------------------------------------------------------------------------*/
#if 1 /* for release tar ball, no relative directory should be existed */
	#include "x_time_msrt.h"
#else /* ori */ 
	#include "time_msrt/x_time_msrt.h"
#endif

/*-----------------------------------------------------------------------------
                    macros, defines, typedefs, enums
 ----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
                    data declarations
 ----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
                    functions declaraions
 ----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
                    Internal functions implementations
 ----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
                    External functions implementations
 ----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
 * Name: c_time_msrt_spot
 *
 * Description: This function create the SPOT TMS record.
 *
 * Inputs:  s_name      Name of the SPOT TMS record.
 *          ui2_tms_lvl The TMS level of the record.
 *
 * Outputs: - 
 *
 * Returns: TMSR_FAIL                 Success.
 *          TMSR_FAIL                 Fail.
 *          TMSR_INSUFFICIENT_MEMORY  Fail.
 *
 ----------------------------------------------------------------------------*/
INT32 c_time_msrt_spot( 
    const CHAR*           s_name,
    UINT16                ui2_tms_lvl,
    const CHAR*           s_comment )
{
    return x_time_msrt_spot( s_name, ui2_tms_lvl, s_comment );
}

/*-----------------------------------------------------------------------------
 * Name: c_time_msrt_diff
 *
 * Description: This function create the SPOT TMS record.
 *
 * Inputs:  s_name      Name of the SPOT TMS record.
 *          ui2_tms_lvl The TMS level of the record.
 *
 * Outputs: - 
 *
 * Returns: TMSR_FAIL                 Success.
 *          TMSR_FAIL                 Fail.
 *          TMSR_INSUFFICIENT_MEMORY  Fail.
 *
 ----------------------------------------------------------------------------*/
INT32 c_time_msrt_diff( 
    const CHAR*           s_name,
    UINT16                ui2_tms_lvl,
    const CHAR*           s_comment )
{
    return x_time_msrt_diff( s_name, ui2_tms_lvl, s_comment );
}

/*-----------------------------------------------------------------------------
 * Name: c_time_msrt_begin
 *
 * Description: This function begins the BEGIN-END TMS record.
 *
 * Inputs:  s_name      Name of the SPOT TMS record.
 *          ui2_avg_cnt The MAX count of the TMS records.
 *          ui2_tms_lvl The TMS level of the record.
 *
 * Outputs: - 
 *
 * Returns: TMSR_FAIL                 Success.
 *          TMSR_FAIL                 Fail.
 *          TMSR_INSUFFICIENT_MEMORY  Fail.
 *
 ----------------------------------------------------------------------------*/
INT32 c_time_msrt_begin( 
    const CHAR*           s_name,
    UINT16                ui2_tms_lvl,
    const CHAR*           s_comment )
{
    return x_time_msrt_begin( s_name, ui2_tms_lvl, s_comment );
}

/*-----------------------------------------------------------------------------
 * Name: c_time_msrt_end
 *
 * Description: This function ends the BEGIN-END TMS record.
 *
 * Inputs:  s_name      Name of the SPOT TMS record.
 *          ui2_tms_lvl The TMS level of the record.
 *
 * Outputs: - 
 *
 * Returns: TMSR_FAIL                 Success.
 *          TMSR_FAIL                 Fail.
 *          TMSR_INSUFFICIENT_MEMORY  Fail.
 *
 ----------------------------------------------------------------------------*/
INT32 c_time_msrt_end( 
    const CHAR*           s_name,
    UINT16                ui2_tms_lvl,
    const CHAR*           s_comment )
{
    return x_time_msrt_end( s_name, ui2_tms_lvl, s_comment );
}

/*-----------------------------------------------------------------------------
 * Name: c_time_msrt_dump
 *
 * Description: This function dumps all records which matchs the name specified
 *              by caller.
 *
 * Inputs:  s_name Caller specified record name to be dump. Set it to NULL to 
 *                 dump all.
 *
 * Outputs: - 
 *
 * Returns: -
 *
 ----------------------------------------------------------------------------*/
VOID c_time_msrt_dump( const CHAR* s_name )
{
    x_time_msrt_dump(s_name);
}

/*-----------------------------------------------------------------------------
 * Name: c_time_msrt_dump_detail
 *
 * Description: This function dumps details of all records which matchs the 
 *              name specified by caller.
 *
 * Inputs:  s_name Caller specified record name to be dump. Set it to NULL to 
 *                 dump all.
 *
 * Outputs: - 
 *
 * Returns: -
 *
 ----------------------------------------------------------------------------*/
VOID c_time_msrt_dump_detail( const CHAR* s_name )
{
    x_time_msrt_dump_detail(s_name);
}

/*-----------------------------------------------------------------------------
 * Name: c_time_msrt_clean
 *
 * Description: This function clean up all records which matchs the 
 *              name specified by caller.
 *
 * Inputs:  s_name Caller specified record name to be clean. Set it to NULL to 
 *                 clean all.
 *
 * Outputs: - 
 *
 * Returns: -
 *
 ----------------------------------------------------------------------------*/
VOID c_time_msrt_clean( const CHAR* s_name )
{
    x_time_msrt_clean(s_name);
}
