/* SPDX-License-Identifier: LGPL-2.1-only OR BSD-3-Clause */
/******************************************************************************
 *
 * This file is provided under a dual license.  When you use or
 * distribute this software, you may choose to be licensed under
 * version 2.1 of the GNU Lesser General Public License
 * ("LGPLv2.1 License") or BSD License.
 *
 * LGPLv2.1 License
 *
 * Copyright(C) 2019 MediaTek Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2.1 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See https://www.gnu.org/licenses/old-licenses/lgpl-2.1.html for more
 * details.
 *
 * BSD LICENSE
 *
 * Copyright(C) 2019 MediaTek Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *  * Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *****************************************************************************/

/*-----------------------------------------------------------------------------
 *
 * $Author: weibo.li $
 * $Date: 2016/02/23 $
 * $RCSfile: cb_low.c,v $
 * $Revision: #21 $
 *
 *---------------------------------------------------------------------------*/

/** @file cb_low.c
 *  This file implements low-layer support of callback interface.
 */


//-----------------------------------------------------------------------------
// Include files
//-----------------------------------------------------------------------------

#include "cb_tim.h"
#include "cb_low.h"
#include "x_linux.h"
#include "x_assert.h"
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/wait.h>
#include <linux/sched.h>
#include <linux/version.h>

//-----------------------------------------------------------------------------
// Static variables
//-----------------------------------------------------------------------------
#define CC_MTAL_CB_REGISTRY_INFO_DUMP

DECLARE_WAIT_QUEUE_HEAD(_rWq);
static DLIST_T(_CB_CALLBACK_EVENT_T) _qCbEvent;

/* Event type to callback thread ownership array */
static pid_t _arCbId[CB_FCT_NUM];
#ifdef CC_MTAL_MULTIPROCESS_CALLBACK
static pid_t _arMultiCbId[CB_FCT_NUM][CB_MAX_REG_NUM];
#endif
static BOOL fgInit=FALSE;
static int iCbCount=0;

#ifdef CC_MTAL_CB_REGISTRY_INFO_DUMP
static CB_REGISTRY_INFO_T *prCBRegDbgInfo = NULL;
static CB_REGISTRY_INFO_T *prCBUnRegDbgInfo = NULL;
static CB_REGISTRY_INFO_T *prCBReleaseDbgInfo = NULL;
#endif

static CB_FCT_ID_T CbDisplayFctIdMin=CB_TIM_TRIGGER;
static CB_FCT_ID_T CbDisplayFctIdMax=CB_FCT_NUM;
static BOOL fgCbDisplayFctId=FALSE;

static int _arCbFctId[CB_FCT_NUM];
#ifdef CC_MTAL_MULTIPROCESS_CALLBACK
static int _arMultiCbFctId[CB_FCT_NUM][3];
#endif

// This code is defined in kernel. Can be default 0. Can be changed by
// 1. uboot CMD: addboot cb_dbg_msk=xxx
// 2. Driver CLI CMD: linux.cb_dbg_msk xxx
extern unsigned int cb_dbg_msk;
#define CB_DBG_GET_EVENT_FLG        (1 << 0)        // Control DBG_Printf function open/close
#define CB_DBG_REG_EVENT_RECORD     (1 << 1)        // Control CB event record or not
#define DBG_Printf(flag, fmt...)    {if(cb_dbg_msk&flag) printk(fmt);}

//-----------------------------------------------------------------------------
// Inter-file functions
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Static functions
//-----------------------------------------------------------------------------
#ifdef CC_MTAL_CB_REGISTRY_INFO_DUMP
static BOOL _CB_doCbRegRecord(VOID)
{
    if (((CB_DBG_REG_EVENT_RECORD&cb_dbg_msk)!=0) &&
        ((prCBRegDbgInfo==NULL)||(prCBUnRegDbgInfo==NULL)||(prCBReleaseDbgInfo==NULL)))
    {
        if (prCBRegDbgInfo == NULL)
        {
            prCBRegDbgInfo = x_mem_alloc(sizeof(CB_REGISTRY_INFO_T)*CB_REGISTRY_INFO_BUFFER_SIZE);
        }

        if (prCBUnRegDbgInfo == NULL)
        {
            prCBUnRegDbgInfo = x_mem_alloc(sizeof(CB_REGISTRY_INFO_T)*CB_REGISTRY_INFO_BUFFER_SIZE);
        }

        if (prCBReleaseDbgInfo == NULL)
        {
            prCBReleaseDbgInfo = x_mem_alloc(sizeof(CB_REGISTRY_INFO_T)*CB_REGISTRY_INFO_BUFFER_SIZE);
        }

        if ((prCBRegDbgInfo==NULL) || (prCBUnRegDbgInfo==NULL) || (prCBReleaseDbgInfo==NULL))
        {
            ASSERT(0);
            if (prCBRegDbgInfo!=NULL)
            {
                x_mem_free(prCBRegDbgInfo);
                prCBRegDbgInfo = NULL;
            }

            if (prCBUnRegDbgInfo!=NULL)
            {
                x_mem_free(prCBUnRegDbgInfo);
                prCBUnRegDbgInfo = NULL;
            }

            if (prCBReleaseDbgInfo!=NULL)
            {
                x_mem_free(prCBReleaseDbgInfo);
                prCBReleaseDbgInfo = NULL;
            }

            return FALSE;
        }

        x_memset(prCBRegDbgInfo, 0, sizeof(CB_REGISTRY_INFO_T)*CB_REGISTRY_INFO_BUFFER_SIZE);
        x_memset(prCBUnRegDbgInfo, 0, sizeof(CB_REGISTRY_INFO_T)*CB_REGISTRY_INFO_BUFFER_SIZE);
        x_memset(prCBReleaseDbgInfo, 0, sizeof(CB_REGISTRY_INFO_T)*CB_REGISTRY_INFO_BUFFER_SIZE);
    }

    return ((CB_DBG_REG_EVENT_RECORD&cb_dbg_msk) != 0);
}
#else
static _CB_doCbRegRecord(void)    (FALSE)
#endif

#ifdef CC_MTAL_CB_REGISTRY_INFO_DUMP
#ifdef CC_MTAL_MULTIPROCESS_CALLBACK
static VOID _CB_DbgCurIDDump(VOID)
{
    struct task_struct *pTask=NULL, *pTmp=NULL;
    BOOL fgFound=FALSE;
    INT32 i4FctIdx, i4FctSubInx;

    if (!_CB_doCbRegRecord()) return;
    printk("[_CB_ParsingEventQueue]====CB Current ID Info (Multi CB)====\n");
    for (i4FctIdx=0; i4FctIdx<CB_FCT_NUM; i4FctIdx++)
    {
        for (i4FctSubInx=0; i4FctSubInx<CB_MAX_REG_NUM; i4FctSubInx++)
        {
            fgFound = FALSE;
            if (_arMultiCbId[i4FctIdx][i4FctSubInx])
            {
                do_each_thread(pTask, pTmp)
                {
                    if (pTmp->pid == _arMultiCbId[i4FctIdx][i4FctSubInx])
                    {
                        printk("[CB_DBG] MCBID[%d][%d] now is grasped by (%d)(%s)\n",
                                i4FctIdx, i4FctSubInx, pTmp->pid, pTmp->comm);
                        fgFound = TRUE;

                        break;
                    }
                } while_each_thread(pTask, pTmp);

                if (!fgFound)
                {
                    printk("[CB_DBG] MCBID[%d][%d] become unhandled by (%d)!!\n",
                            i4FctIdx, i4FctSubInx, _arMultiCbId[i4FctIdx][i4FctSubInx]);
                }
            }
        }
    }

    return ;
}
#else
static VOID _CB_DbgCurIDDump(VOID)
{
    struct task_struct *pTask=NULL, *pTmp=NULL;
    BOOL fgFound=FALSE;
    INT32 i4Count;

    if (!_CB_doCbRegRecord()) return;
    printk("[_CB_ParsingEventQueue]====CB Current ID Info (Single CB)====\n");
    for (i4Count=0; i4Count<CB_FCT_NUM; i4Count++)
    {
        fgFound = FALSE;
        if (_arCbId[i4Count])
        {
            do_each_thread(pTask, pTmp)
            {
                if (pTmp->pid == _arCbId[i4Count])
                {
                    printk("[CB_DBG] CBID (%d) now is grasped by (%d)(%s)\n",
                            i4Count, pTmp->pid, pTmp->comm);
                            fgFound = TRUE;
                    break;
                }
            } while_each_thread(pTask, pTmp);

            if (!fgFound)
            {
                printk("[CB_DBG] CBID (%d) become unhandled by (%d)!!\n",
                    i4Count, _arCbId[i4Count]);
            }
        }
    }

    return ;
}
#endif

static VOID _CB_DbgRegRec(INT32 i4Pid, INT32 i4FctID)
{
    INT32 i4Count;

    if (!_CB_doCbRegRecord()) return;
    for (i4Count=0; i4Count<CB_REGISTRY_INFO_BUFFER_SIZE; i4Count++)
    {
        if (prCBRegDbgInfo && !prCBRegDbgInfo[i4Count].fgRegistered)
        {
            prCBRegDbgInfo[i4Count].i4RegistryPid = i4Pid;
            prCBRegDbgInfo[i4Count].i4RegistryFctId = i4FctID;
            prCBRegDbgInfo[i4Count].fgRegistered = TRUE;

            break;
        }
    }

    return ;
}

static VOID _CB_DbgRegDump(VOID)
{
    struct task_struct *pTask=NULL, *pTmp=NULL;
    INT32 i4Count;
    BOOL fgFound=FALSE;

    if (!_CB_doCbRegRecord()) return;
    printk("[_CB_ParsingEventQueue]====CB Registery Record====\n");
    for (i4Count=0; i4Count<CB_REGISTRY_INFO_BUFFER_SIZE; i4Count++)
    {
        fgFound = FALSE;

        if (prCBRegDbgInfo[i4Count].fgRegistered)
        {
            do_each_thread(pTask, pTmp)
            {
                if(pTmp->pid == prCBRegDbgInfo[i4Count].i4RegistryPid)
                {
                    printk("[CB_DBG] CBID (%d) was registered by (%d)(%s)\n",
                    prCBRegDbgInfo[i4Count].i4RegistryFctId,
                    pTmp->pid, pTmp->comm);
                    fgFound = TRUE;

                    break;
                }
            } while_each_thread(pTask, pTmp);

            if (!fgFound)
            {
              printk("[CB_DBG] CBID (%d) was abandoned (%d)!!\n",
                     prCBRegDbgInfo[i4Count].i4RegistryFctId,
                     prCBRegDbgInfo[i4Count].i4RegistryPid);
            }
        }
    }

    return ;
}

static VOID _CB_DbgUnregRec(INT32 i4Pid, INT32 i4FctID)
{
    INT32 i4Count;

    if (!_CB_doCbRegRecord()) return;
    for (i4Count=0; i4Count<CB_REGISTRY_INFO_BUFFER_SIZE; i4Count++)
    {
        if (prCBUnRegDbgInfo && !prCBUnRegDbgInfo[i4Count].fgRegistered)
        {
            prCBUnRegDbgInfo[i4Count].i4RegistryPid = i4Pid;
            prCBUnRegDbgInfo[i4Count].i4RegistryFctId = i4FctID;
            prCBUnRegDbgInfo[i4Count].fgRegistered = TRUE;

            break;
        }
    }

    return ;
}

static VOID _CB_DbgUnregDump(VOID)
{
    struct task_struct *pTask=NULL, *pTmp=NULL;
    INT32 i4Count;
    BOOL fgFound=FALSE;

    if (!_CB_doCbRegRecord()) return;
    printk("[_CB_ParsingEventQueue]====CB Un-Registery Record====\n");
    for (i4Count=0; i4Count<CB_REGISTRY_INFO_BUFFER_SIZE;i4Count++)
    {
        fgFound = FALSE;
        if (prCBUnRegDbgInfo[i4Count].fgRegistered)
        {
            do_each_thread(pTask, pTmp)
            {
                if (pTmp->pid == prCBUnRegDbgInfo[i4Count].i4RegistryPid)
                {
                    printk("[CB_DBG] CBID (%d) was un-registered by (%d)(%s)\n",
                            prCBUnRegDbgInfo[i4Count].i4RegistryFctId,
                            pTmp->pid, pTmp->comm);
                    fgFound = TRUE;
                    break;
                }
            } while_each_thread(pTask, pTmp);

            if (!fgFound)
            {
                printk("[CB_DBG] CBID (%d) was abandoned (%d)!!\n",
                        prCBUnRegDbgInfo[i4Count].i4RegistryFctId,
                        prCBUnRegDbgInfo[i4Count].i4RegistryPid);
            }
        }
    }

    return ;
}

static VOID _CB_DbgReleaseRec(INT32 i4Pid, INT32 i4FctID)
{
    INT32 i4Count;

    if (!_CB_doCbRegRecord()) return;
    for (i4Count=0; i4Count<CB_REGISTRY_INFO_BUFFER_SIZE; i4Count++)
    {
        if (!prCBReleaseDbgInfo[i4Count].fgRegistered)
        {
            prCBReleaseDbgInfo[i4Count].i4RegistryPid = i4Pid;
            prCBReleaseDbgInfo[i4Count].i4RegistryFctId = i4FctID;
            prCBReleaseDbgInfo[i4Count].fgRegistered = TRUE;

            break;
        }
    }

    return ;
}

static VOID _CB_DbgReleaseDump(VOID)
{
    INT32 i4Count;

    if (!_CB_doCbRegRecord()) return;
    printk("[_CB_ParsingEventQueue]====CB Release Record====\n");
    for (i4Count=0; i4Count<CB_REGISTRY_INFO_BUFFER_SIZE; i4Count++)
    {
        if (prCBReleaseDbgInfo[i4Count].fgRegistered)
        {
            printk("[CB_DBG] CBID (%d) was released by (%d)!!\n",
                    prCBReleaseDbgInfo[i4Count].i4RegistryFctId,
                    prCBReleaseDbgInfo[i4Count].i4RegistryPid);
        }
    }

    return ;
}
#else
#define _CB_DbgCurIDDump()
#define _CB_DbgRegInit()
#define _CB_DbgRegRec(a, b)
#define _CB_DbgRegDump()
#define _CB_DbgUnregInit()
#define _CB_DbgUnregRec(a, b)
#define _CB_DbgUnregDump()
#define _CB_DbgReleaseInit()
#define _CB_DbgReleaseRec(a, b)
#define _CB_DbgReleaseDump()
#endif

//-----------------------------------------------------------------------------
/** Add callback event into event queue.
 *  @param  prCbEv  Pointer to callback event.
 *  @retval CBR_OK          Success.
 *  @retval CBR_ERR_SYS     Internal system error.
 */
//-----------------------------------------------------------------------------
static INT32 _CbPutEvent(CB_CALLBACK_EVENT_T *prCbEv)
{
    CRIT_STATE_T rState;

    rState = x_crit_start();
    DLIST_INSERT_TAIL(prCbEv, &_qCbEvent, rLink);
    iCbCount++;
    _arCbFctId[prCbEv->rGetCb.eFctId]++;
    #ifdef CC_MTAL_MULTIPROCESS_CALLBACK
    if(prCbEv->u4BitMask & BITMASK_MULTI1)
    {
      _arMultiCbFctId[prCbEv->rGetCb.eFctId][0]++;
    }
    if(prCbEv->u4BitMask & BITMASK_MULTI2)
    {
      _arMultiCbFctId[prCbEv->rGetCb.eFctId][1]++;
    }
    if(prCbEv->u4BitMask & BITMASK_MULTI3)
    {
      _arMultiCbFctId[prCbEv->rGetCb.eFctId][2]++;
    }
    #endif
    x_crit_end(rState);

    wake_up_interruptible_all(&_rWq);

    return CBR_OK;
}

//-----------------------------------------------------------------------------
/** Initialize callback support.
 */
//-----------------------------------------------------------------------------
#if defined(CC_FAST_INIT)
extern int  mt53xx_drvset_init(void);
#endif
void _CB_Init(void)
{
    CRIT_STATE_T rState;
    UINT32 ui4Idx;

    UNUSED(ui4Idx);
    #if defined(CC_FAST_INIT)
    mt53xx_drvset_init();
    #endif
    #if defined(CC_SUPPORT_STR)
    mt53xx_STR_init();
    #endif
    if (fgInit)
    {
        return;
    }

    rState = x_crit_start();
    fgInit = TRUE;
    DLIST_INIT(&_qCbEvent);
    iCbCount = 0;
    for(ui4Idx=0;ui4Idx<CB_FCT_NUM;ui4Idx++)
    {
        _arCbFctId[ui4Idx] = 0;
#ifdef CC_MTAL_MULTIPROCESS_CALLBACK
        _arMultiCbFctId[ui4Idx][0] = 0;
        _arMultiCbFctId[ui4Idx][1] = 0;
        _arMultiCbFctId[ui4Idx][2] = 0;
#endif
    }

    x_crit_end(rState);

    UNUSED(_CB_doCbRegRecord());

    return ;
}

//-----------------------------------------------------------------------------
/** To enable/register a callback function with the related ID.
 *  @param  eFctId  Callback function ID to enable.
 *  @param  i4SubId Callback function Sub ID.
 *  @return The return value of related callback registration function.
 */
//-----------------------------------------------------------------------------
INT32 _CB_RegCbFct(CB_FCT_ID_T eFctId, INT32 i4SubId, pid_t rCbPid)
{
    BOOL fgWakeUp=FALSE;
    CRIT_STATE_T rState;
    BOOL fgSCBypass=TRUE;
    struct task_struct *pTask, *pTmp;
    UINT32 uid;

	if((eFctId < 0) ||  (eFctId >= CB_FCT_NUM))
	{
		printk("[Error] eFctId out of range %d\n", eFctId);
		return CBR_ERR_REG;
	}
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 9, 0)
    uid = current_uid();
#else
    uid = current_uid().val;
#endif

    if(eFctId == CB_SC_TRIGGER && uid != 0){
        /* non root process ignore SC_TRIGGER register callback */
        printk("$$$ Thread PID not root(%u) ignore\n", uid);
        return CBR_OK;
    }
    if ((eFctId==CB_SC_TRIGGER) && (_arCbId[CB_SC_TRIGGER]!=(pid_t)0))
    {
        printk("$$$ Try to register CB_SC_TRIGGER rCbPid=%d, current pid=%d, current comm=%s\n",
                rCbPid, current->pid, current->comm);
          do_each_thread(pTask, pTmp)
          {
            if(pTmp->pid == rCbPid)
            {
              printk("_CB_RegCbFct[CB_SC_TRIGGER RESET] set by (%d)(%s)\n",pTmp->pid, pTmp->comm);
              if(strncmp(pTmp->comm,"dtv_svc",strlen(pTmp->comm))==0)
              {
                printk("dtv_svc grasped CB_SC_TRIGGER_RESET\n");
                fgSCBypass = FALSE;
              }
                  break;
            }
          } while_each_thread(pTask, pTmp);
    }
    if((eFctId==CB_SC_TRIGGER) && fgSCBypass && (_arCbId[CB_SC_TRIGGER]!=(pid_t)0)) {
        printk("$$$ _CB_RegCbFct[CB SC] pid(%d) skip set by pid(%d)\n",
                (INT32)rCbPid, (INT32)_arCbId[CB_SC_TRIGGER]);
        return CBR_OK;
    }

    #if 0
    if ((_arMultiCbId[eFctId][0] != 0) ||
        (_arMultiCbId[eFctId][1] != 0) ||
        (_arMultiCbId[eFctId][2] != 0))
    {
        return CBR_ERR_REG;
    }
    #endif
    rState = x_crit_start();
    if (eFctId < CB_FCT_NUM)
    {
        if ((_arCbId[eFctId] != (pid_t)0) &&
            (_arCbId[eFctId] != rCbPid))
        {
            fgWakeUp = TRUE;
        }

        _CB_DbgRegRec((INT32)current->pid,(INT32)eFctId);
        _arCbId[eFctId] = rCbPid;
    }
    x_crit_end(rState);

#ifndef CONFIG_TV_DRV_VFY
    if (apfnCbRegIf[eFctId] != NULL)
    {
        return apfnCbRegIf[eFctId](i4SubId);
    }
#endif
    if (fgWakeUp)
    {
        wake_up_interruptible_all(&_rWq);
    }

    /*
    printk("*** Thread 0x%x registers callback (0x%x, 0x%x)\n",
           (INT32)current, (INT32)eFctId, (INT32)i4SubId);
    */

    return CBR_OK;
}

#ifdef CC_MTAL_MULTIPROCESS_CALLBACK
//-----------------------------------------------------------------------------
/** To enable/register a multiple callback function with the related ID.
 *  @param  eFctId  Callback function ID to enable.
 *  @param  i4SubId Callback function Sub ID.
 *  @return The return value of related multiple callback registration function.
 */
//-----------------------------------------------------------------------------
INT32 _CB_RegMultiCbFct(CB_FCT_ID_T eFctId, INT32 i4SubId, pid_t rCbPid)
{
    struct task_struct *pTask, *pTmp;
    BOOL fgSCBypass=TRUE;
    BOOL fgWakeUp=FALSE;
    CRIT_STATE_T rState;
    INT32 i4Ret = 0;
    UINT32 uid;

	if((eFctId < 0) ||  (eFctId >= CB_FCT_NUM))
	{
		printk("[Error] eFctId out of range %d\n", eFctId);
		return CBR_ERR_REG;
	}

#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 9, 0)
    uid = current_uid();
#else
    uid = current_uid().val;
#endif
    if(eFctId == CB_SC_TRIGGER && uid != 0){
        /* non root process ignore SC_TRIGGER register callback */
        printk("$$$ Thread PID not root(%u) ignore\n", uid);
        return CBR_OK;
    }

    if((eFctId==CB_SC_TRIGGER) && (_arMultiCbId[CB_SC_TRIGGER][0]!=0))
    {
      printk("$$$ Try to register CB_SC_TRIGGER rCbPid=%d, current pid=%d, current comm=%s\n",
             rCbPid, current->pid, current->comm);
      do_each_thread(pTask, pTmp)
      {
        if(pTmp->pid == rCbPid)
        {
          printk("_CB_RegMultiCbFct[CB_SC_TRIGGER RESET] set by (%d)(%s)\n",pTmp->pid, pTmp->comm);
          if(strncmp(pTmp->comm,"dtv_svc",strlen(pTmp->comm))==0)
          {
            printk("dtv_svc grasped CB_SC_TRIGGER_RESET\n");
            fgSCBypass = FALSE;
          }
          break;
        }
      } while_each_thread(pTask, pTmp);
    }

    if((eFctId==CB_SC_TRIGGER) && fgSCBypass && (_arMultiCbId[CB_SC_TRIGGER][0]!=0)){
        printk("$$$ _CB_RegMultiCbFct[CB SC] pid(%d) skip set by pid(%d)\n",
                (INT32)rCbPid, (INT32)_arMultiCbId[CB_SC_TRIGGER][0]);
        return CBR_OK;
    }

    rState = x_crit_start();
    if (eFctId < CB_FCT_NUM)
    {
        if ((_arCbId[eFctId] != (pid_t)0) &&
            (_arCbId[eFctId] != rCbPid))
        {
            fgWakeUp = TRUE;    // wakeup to leave.
        }

        _arCbId[eFctId] = 0;    // Remove single callback.
        if (_arMultiCbId[eFctId][0] == 0)
        {
            _arMultiCbId[eFctId][0] = rCbPid;
            _CB_DbgRegRec((INT32)current->pid, (INT32)eFctId);
        }
        else if (_arMultiCbId[eFctId][1] == 0)
        {
            _arMultiCbId[eFctId][1] = rCbPid;
            _CB_DbgRegRec((INT32)current->pid, (INT32)eFctId);
        }
        else if (_arMultiCbId[eFctId][2] == 0)
        {
            _arMultiCbId[eFctId][2] = rCbPid;
            _CB_DbgRegRec((INT32)current->pid, (INT32)eFctId);
        }
        else
        {
            #if CB_MULTIPROCESS_LAST_REPLACE
            if((rCbPid != _arMultiCbId[eFctId][0]) && (rCbPid != _arMultiCbId[eFctId][1]) &&
                (rCbPid != _arMultiCbId[eFctId][2]))
            {
                _arMultiCbId[eFctId][2] = rCbPid;
                _CB_DbgRegRec((INT32)current->pid, (INT32)eFctId);
            }
            else
            {
                printk("_CB_RegMultiCbFct multi-fctid used in process(%d)\n",rCbPid);
            }
            #else
            i4Ret = CBR_ERR_REG;
            #endif
        }
        _arCbId[eFctId] = (pid_t)0;
    }
    x_crit_end(rState);

    if (i4Ret) return i4Ret;

#ifndef CONFIG_TV_DRV_VFY
    if (apfnCbRegIf[eFctId] != NULL)
    {
        return apfnCbRegIf[eFctId](i4SubId);
    }
#endif
    if (fgWakeUp)
    {
        wake_up_interruptible_all(&_rWq);
    }

    /*
    printk("*** Thread 0x%x registers multiple callback (0x%x, 0x%x)\n",
           (INT32)current, (INT32)eFctId, (INT32)i4SubId);
    */

    return CBR_OK;
}
#endif

pid_t _CB_GetPid(CB_FCT_ID_T eFctId)
{
    if(eFctId < CB_FCT_NUM)
    {
        #ifdef CC_MTAL_MULTIPROCESS_CALLBACK
        if ((current->tgid == _arCbId[eFctId]) ||
            (current->tgid == _arMultiCbId[eFctId][0]) ||
            (current->tgid == _arMultiCbId[eFctId][1]) ||
            (current->tgid == _arMultiCbId[eFctId][2]))
        {
            return current->tgid;
        }

        if (_arCbId[eFctId]) return _arCbId[eFctId];
        if (_arMultiCbId[eFctId][0]) return _arMultiCbId[eFctId][0];
        if (_arMultiCbId[eFctId][1]) return _arMultiCbId[eFctId][1];
        if (_arMultiCbId[eFctId][2]) return _arMultiCbId[eFctId][2];
        #endif
        return _arCbId[eFctId];
    }
    else
    {
        return 0;
    }
}

BOOL _CB_ParsingEventQueue(CB_EVENT_COUNT_T *prCbIdleFctIdInfo)
{
    CB_CALLBACK_EVENT_T *prLast=NULL;
    CB_FCT_ID_T eType;
    CRIT_STATE_T rState;

    if (prCbIdleFctIdInfo == NULL)
    {
        return FALSE;
    }

    for(eType=CB_TIM_TRIGGER;eType<CB_FCT_NUM;eType++)
    {
        prCbIdleFctIdInfo->puCB00IdleFctPid[eType]=0;
#ifdef CC_MTAL_MULTIPROCESS_CALLBACK
        prCbIdleFctIdInfo->puCB01IdleFctPid[eType]=0;
        prCbIdleFctIdInfo->puCB02IdleFctPid[eType]=0;
#endif
    }

    _CB_DbgRegDump();
    _CB_DbgUnregDump();
    _CB_DbgReleaseDump();
    _CB_DbgCurIDDump();

    rState = x_crit_start();
    DLIST_FOR_EACH(prLast, &_qCbEvent, rLink)
    {
        if (!prLast)
        {
            break;
        }

        if ((prLast->rGetCb.eFctId>=CB_FCT_NUM) || (prLast->rGetCb.eFctId<CB_TIM_TRIGGER))
        {
            break;
        }

        prCbIdleFctIdInfo->puCB00IdleFctPid[prLast->rGetCb.eFctId]++;
#ifdef CC_MTAL_MULTIPROCESS_CALLBACK
        prCbIdleFctIdInfo->puCB01IdleFctPid[prLast->rGetCb.eFctId]++;
        prCbIdleFctIdInfo->puCB02IdleFctPid[prLast->rGetCb.eFctId]++;
#endif
    }
    x_crit_end(rState);

    return TRUE;
}

int _CB_GetCbCount(void)
{
    CRIT_STATE_T rState;
    int i4Count;
    CB_FCT_ID_T Fctidx;
    //printk("_CB_GetCbCount = %d!\n",iCbCount);
    rState = x_crit_start();
    i4Count = iCbCount;
    x_crit_end(rState);

    printk("[%s]====dump CB count list====\n",__FUNCTION__);
    for (Fctidx=CB_TIM_TRIGGER; Fctidx<CB_FCT_NUM; Fctidx++)
    {
#ifdef CC_MTAL_MULTIPROCESS_CALLBACK
        printk("CBID[%d] count(%d)(%d %d %d)\n", Fctidx, _arCbFctId[Fctidx],
                _arMultiCbFctId[Fctidx][0], _arMultiCbFctId[Fctidx][1],
                _arMultiCbFctId[Fctidx][2]);
#else
        printk("CBID[%d] count(%d)\n", Fctidx, _arCbFctId[Fctidx]);
#endif
    }

    return i4Count;
}

void _CB_GetCbIdDisplayInfo(void)
{
  printk("[%s]--->\n",__FUNCTION__);
  printk("fgCbDisplayFctId = %d\n",fgCbDisplayFctId);
  printk("CbDisplayFctId(Min/Max) = %d / %d\n",CbDisplayFctIdMin,CbDisplayFctIdMax);

  return ;
}

int _CB_SetCbIdDisplayInfo(BOOL fgDisplay, CB_FCT_ID_T Min, CB_FCT_ID_T Max)
{
  int i4Ret = -1;

  printk("[%s]\n",__FUNCTION__);
  fgCbDisplayFctId = fgDisplay;
  if((Max > CB_FCT_NUM) || (Min > Max))
  {
    printk("setting error and disable display!!\n");
    fgCbDisplayFctId = FALSE;
  }
  else
  {
    i4Ret = 0;
    CbDisplayFctIdMin = Min;
    CbDisplayFctIdMax = Max;
  }
  return i4Ret;
}

pid_t _CB_ParsingSCInfo(void)
{
    struct task_struct *pTask, *pTmp;
    CB_CALLBACK_EVENT_T *prLast=NULL;
    pid_t rSCpid=0;
    UINT32 u4UnhandleEventCount=0;
    BOOL bExist=FALSE;
    CRIT_STATE_T rState;

    #ifdef CC_MTAL_MULTIPROCESS_CALLBACK
    rSCpid = _arMultiCbId[CB_SC_TRIGGER][0];
    #else
    rSCpid = _arCbId[CB_SC_TRIGGER];
    #endif

    do_each_thread(pTask, pTmp)
    {
      if(pTmp->pid == rSCpid)
      {
        printk("[_CB_ParsingSCInfo] SC CB id was grasped by (%d)(%s)\n",pTmp->pid, pTmp->comm);
        bExist = TRUE;
        break;
      }
    } while_each_thread(pTask, pTmp);

    rState = x_crit_start();
    DLIST_FOR_EACH(prLast, &_qCbEvent, rLink)
    {
      if (!prLast)
      {
        break;
      }
      if (prLast->rGetCb.eFctId >= CB_FCT_NUM)
      {
        break;
      }
      if (prLast->rGetCb.eFctId == CB_SC_TRIGGER)
      {
         u4UnhandleEventCount++;
      }
    }
    x_crit_end(rState);
    if(!bExist)
    {
      printk("[_CB_ParsingSCInfo] SC CB handler does not exist or has been lost\n");
      rSCpid = 0;
    }
    printk("[_CB_ParsingSCInfo] SC CB id unhandled count = %d\n",u4UnhandleEventCount);
    return rSCpid;
}

#ifdef CC_MTAL_MULTIPROCESS_CALLBACK
//-----------------------------------------------------------------------------
/** To unregister a callback function with the related ID.
 *  @param  eFctId  Callback function ID to enable.
 *  @param  i4SubId Callback function Sub ID.
 *  @return The return value of related callback unregistration function.
 */
//-----------------------------------------------------------------------------
INT32 _CB_UnRegCbFct(CB_FCT_ID_T eFctId, INT32 i4SubId, pid_t rCbPid)
{
    CB_CALLBACK_EVENT_T *prLast=NULL,tmpCBEvt;
    CRIT_STATE_T rState;

    rState = x_crit_start();
    if ((eFctId < CB_FCT_NUM) && (rCbPid == _arCbId[eFctId]))
    {
        _arCbId[eFctId] = 0;
    }
    DLIST_FOR_EACH(prLast, &_qCbEvent, rLink)
    {
        if ((rCbPid == _arCbId[prLast->rGetCb.eFctId]) &&
            (prLast->rGetCb.eFctId == eFctId))
        {
            prLast->u4BitMask &= ~(BITMASK_SINGLE);
            if ((prLast->u4BitMask == 0)
#ifdef CC_MTAL_CB_REMOVE_LIST_FIXED
              && (prLast->u4RefMask == 0)
#endif
                )
            {
                DLIST_REMOVE(prLast, &_qCbEvent, rLink);
                memcpy(&tmpCBEvt, prLast, sizeof(CB_CALLBACK_EVENT_T));
                kfree(prLast);
                prLast = &tmpCBEvt ;
            }
        }
    }

    x_crit_end(rState);

    return CBR_OK;
}

//-----------------------------------------------------------------------------
/** To unregister a multiple callback function with the related ID.
 *  @param  eFctId  Callback function ID to enable.
 *  @param  i4SubId Callback function Sub ID.
 *  @return The return value of related callback unregistration function.
 */
//-----------------------------------------------------------------------------
INT32 _CB_UnRegMultiCbFct(CB_FCT_ID_T eFctId, INT32 i4SubId, pid_t rCbPid)
{
    CB_CALLBACK_EVENT_T *prLast=NULL, tmpCBEvt;
    CRIT_STATE_T rState;
    UINT32 u4Multi = 0;

    rState = x_crit_start();
    if (eFctId < CB_FCT_NUM)
    {
        if (rCbPid == _arMultiCbId[eFctId][0])
        {
            _arMultiCbId[eFctId][0] = 0;
            u4Multi = 1;
            _CB_DbgUnregRec((INT32)current->pid, (INT32)eFctId);
        }
        if (rCbPid == _arMultiCbId[eFctId][1])
        {
            _arMultiCbId[eFctId][1] = 0;
            u4Multi = 2;
            _CB_DbgUnregRec((INT32)current->pid, (INT32)eFctId);
        }
        if (rCbPid == _arMultiCbId[eFctId][2])
        {
            _arMultiCbId[eFctId][2] = 0;
            u4Multi = 3;
            _CB_DbgUnregRec((INT32)current->pid, (INT32)eFctId);
        }

        DLIST_FOR_EACH(prLast, &_qCbEvent, rLink)
        {
            if ((u4Multi != 0) &&
                (rCbPid == _arMultiCbId[prLast->rGetCb.eFctId][u4Multi-1]) &&
                (prLast->rGetCb.eFctId == eFctId))
            {
                prLast->u4BitMask &= ~(BITMASK_MULTI(u4Multi));
                if ((prLast->u4BitMask == 0)
#ifdef CC_MTAL_CB_REMOVE_LIST_FIXED
                && (prLast->u4RefMask == 0)
#endif
                    )
                {
                    DLIST_REMOVE(prLast, &_qCbEvent, rLink);
                    memcpy(&tmpCBEvt, prLast, sizeof(CB_CALLBACK_EVENT_T));
                    kfree(prLast);
                    prLast = &tmpCBEvt ;
                }
            }
        }
    }

    x_crit_end(rState);

    return CBR_OK;
}
#else
//-----------------------------------------------------------------------------
/** To unregister a callback function with the related ID.
 *  @param  eFctId  Callback function ID to enable.
 *  @param  i4SubId Callback function Sub ID.
 *  @return The return value of related callback unregistration function.
 */
//-----------------------------------------------------------------------------
INT32 _CB_UnRegCbFct(CB_FCT_ID_T eFctId, INT32 i4SubId, pid_t rCbPid)
{
    CRIT_STATE_T rState;
    CB_CALLBACK_EVENT_T *prLast=NULL,tmpCBEvt;

    if (eFctId >= CB_FCT_NUM)
        return CBR_ERR_REG;

    rState = x_crit_start();
    if (rCbPid != _arCbId[eFctId] || !_arCbId[eFctId])
        goto done;

    _arCbId[eFctId] = 0;
    _CB_DbgUnregRec((INT32)current->pid, (INT32)eFctId);

    // Check if this is on event queue.
    DLIST_FOR_EACH(prLast, &_qCbEvent, rLink)
    {
        if (!prLast)
        {
            break;
        }

        if (prLast->rGetCb.eFctId >= CB_FCT_NUM)
        {
            break;
        }

        if (prLast->rGetCb.eFctId == eFctId)
        {
            DLIST_REMOVE(prLast, &_qCbEvent, rLink);
            //workaround for kfree will write the 0x6b for poison check
            memcpy(&tmpCBEvt, prLast, sizeof(CB_CALLBACK_EVENT_T));
            kfree(prLast);
            prLast = &tmpCBEvt ;
        }

        if (prLast->rLink.pt_next == NULL)
        {
            break;
        }
    }

done:
    x_crit_end(rState);

    return CBR_OK;
}
#endif

//-----------------------------------------------------------------------------
/** Add callback event into event queue.
 *  @param  eFctId      Callback function ID to notify.
 *  @param  i4TagSize   Data size to callback.
 *  @param  pvTag       Data to callback.
 *  @retval CBR_OK          Success.
 *  @retval CBR_NO_MEM      Out of memory.
 */
//-----------------------------------------------------------------------------
INT32 _CB_PutEvent(CB_FCT_ID_T eFctId, INT32 i4TagSize, void *pvTag)
{
    CB_CALLBACK_EVENT_T *prCbEv;
    UINT32 flags;
    #ifdef CC_MTAL_MULTIPROCESS_CALLBACK
    UINT32 u4BitMask=0;
    #endif
    if (!fgInit)
    {
        printk("Error! CB not yet initialized!\n");
        return CBR_ERR_SYS;
    }

    if (in_interrupt())
    {
        flags = GFP_ATOMIC;
    }
    else
    {
        flags = GFP_KERNEL;
    }

    if (sizeof(CB_CALLBACK_EVENT_T)+i4TagSize > CB_MAX_STRUCT_SIZE)
    {
        printk("%s:%d ERROR: Event size too large. %d+%d>%d\n", __FUNCTION__, __LINE__,
                sizeof(CB_CALLBACK_EVENT_T), i4TagSize, CB_MAX_STRUCT_SIZE);
    }
    ASSERT((sizeof(CB_GET_CALLBACK_T) + i4TagSize) <= CB_MAX_STRUCT_SIZE);

    prCbEv = (CB_CALLBACK_EVENT_T *)
             kmalloc(sizeof(CB_CALLBACK_EVENT_T) + i4TagSize, flags);

    if (prCbEv == NULL)
    {
        return CBR_NO_MEM;
    }
    #ifdef CC_MTAL_MULTIPROCESS_CALLBACK
    if (_arCbId[eFctId])
    {
        u4BitMask = BITMASK_SINGLE;
    }
    else
    {
        u4BitMask = (_arMultiCbId[eFctId][0] ? (BITMASK_MULTI1) : 0) |
                    (_arMultiCbId[eFctId][1] ? (BITMASK_MULTI2) : 0) |
                    (_arMultiCbId[eFctId][2] ? (BITMASK_MULTI3) : 0);
    }

    if (!u4BitMask)
    {
#ifdef CC_CB_INIT_PUT_UNREG_EVENT
        printk("$$$ _CB_PutEvent bitmask==0!!! (pid %d,eFctId %d)\n", current->pid, eFctId);
        u4BitMask = BITMASK_MULTI1;
#else
        if (eFctId != CB_SC_TRIGGER)
        {
            //No one register callback , skip put event
            printk("[Error] No one register callback , skip put event!\n");
            kfree(prCbEv);
            return CBR_OK;
        }
        else
        {
            printk("$$$ _CB_PutEvent bitmask==0!!! (pid %d,eFctId %d)\n", current->pid, eFctId);
            u4BitMask = BITMASK_MULTI1;
        }
#endif
    }

    prCbEv->u4BitMask = u4BitMask;
#ifdef CC_MTAL_CB_REMOVE_LIST_FIXED
    prCbEv->u4RefMask = 0;
#endif
    #endif
    if(fgCbDisplayFctId)
    {
      if(eFctId >= CbDisplayFctIdMin && eFctId <= CbDisplayFctIdMax)
      {
        printk("_CB_PutEvent (pid %d,eFctId %d)(Max/Min)\n",
               current->pid, eFctId, CbDisplayFctIdMin, CbDisplayFctIdMax);
      }
    }

    prCbEv->rGetCb.eFctId = eFctId;
    prCbEv->rGetCb.i4TagSize = i4TagSize;
    memcpy((void *)((UINT32)prCbEv + sizeof(CB_CALLBACK_EVENT_T)), pvTag,
           i4TagSize);
    _CbPutEvent(prCbEv);

    return CBR_OK;
}

#define CB_OTHER_OVERFLOW 256
#define SC_OTHER_OVERFLOW 16

#ifdef CC_MTAL_MULTIPROCESS_CALLBACK
//-----------------------------------------------------------------------------
/** Get the first event from event queue.
 *  @return If success and has event in queue, pointer of the callback event.\n
 *          Otherwise, NULL.
 */
//-----------------------------------------------------------------------------
CB_CALLBACK_EVENT_T *_CB_GetEvent(UINT32 *pu4DoneMask)
{
    CB_CALLBACK_EVENT_T *prLast=NULL;
    CRIT_STATE_T rState;
    BOOL fgFound=FALSE;
    UINT32 u4DoneMask = 0;
    static UINT32 cnt_o=0, cnt_sc=0;
    UINT32 i=0, j=0;

    rState = x_crit_start();
    DLIST_FOR_EACH(prLast, &_qCbEvent, rLink)
    {
        if (!prLast)
        {
            break;
        }

        if (prLast->rGetCb.eFctId >= CB_FCT_NUM)
        {
            break;
        }

        if (current->tgid == _arCbId[prLast->rGetCb.eFctId])
        {
            fgFound = TRUE;
            u4DoneMask = BITMASK_SINGLE;
            break;
        }
        else if ((current->tgid == _arMultiCbId[prLast->rGetCb.eFctId][0]) &&
                 (prLast->u4BitMask & BITMASK_MULTI1))
        {
            fgFound = TRUE;
            u4DoneMask = BITMASK_MULTI1;
            break;
        }
        else if ((current->tgid == _arMultiCbId[prLast->rGetCb.eFctId][1]) &&
                 (prLast->u4BitMask & BITMASK_MULTI2))
        {
            fgFound = TRUE;
            u4DoneMask = BITMASK_MULTI2;
            break;
    }
        else if ((current->tgid == _arMultiCbId[prLast->rGetCb.eFctId][2]) &&
                 (prLast->u4BitMask & BITMASK_MULTI3))
    {
            fgFound = TRUE;
            u4DoneMask = BITMASK_MULTI3;
            break;
          }
        i++;
        if (prLast->rGetCb.eFctId == CB_SC_TRIGGER)
        {
            j++;
        }
        }
#ifdef CC_MTAL_CB_REMOVE_LIST_FIXED
    if (pu4DoneMask!=NULL)
    {
        *pu4DoneMask = u4DoneMask;
        //Note: Not fgFound case , should check again
        if(fgFound && (prLast != NULL))
        {
            prLast->u4BitMask &= ~(u4DoneMask);
            prLast->u4RefMask |= u4DoneMask ;
        }
    }
    x_crit_end(rState);
#else
    x_crit_end(rState);
    if (pu4DoneMask!=NULL)
    {
        *pu4DoneMask = u4DoneMask;
    }
#endif
    cnt_o += i;
    cnt_sc += j;


    if (fgFound)
    {
#ifdef CC_MTAL_CB_REMOVE_LIST_FIXED
        DBG_Printf(CB_DBG_GET_EVENT_FLG, "%s:pid %d/%d prLast %x u4DoneMask %d , u4BitMask %d u4RefMask %d fctid %d\n", __FUNCTION__,
                   current->tgid, current->pid,  prLast, u4DoneMask , prLast->u4BitMask, prLast->u4RefMask, prLast->rGetCb.eFctId);
#endif
        return prLast;
    }
    else
    {
        if (i == 0)
        {
            cnt_o = 0;
        }
        if (cnt_o > CB_OTHER_OVERFLOW)
        {
            if(_CB_doCbRegRecord())
            {
                printk("[Error] cb events of others overflow\n");
            }
        }
        if (j == 0)
        {
            cnt_sc = 0;
        }
        if (cnt_sc > SC_OTHER_OVERFLOW)
        {
            if(_CB_doCbRegRecord())
            {
                printk("[Error] cb events of sc overflow\n");
            }
        }
        return NULL;
    }
}
//-----------------------------------------------------------------------------
/** Free the element.
 *  @param  prCbEv      event element pointer to be removed and free.
 *  @param  u4DoneMask  input parameter to identify the Done Mask bit.
 */
//-----------------------------------------------------------------------------
void _CB_FreeEvent(CB_CALLBACK_EVENT_T *prCbEv, UINT32 u4DoneMask)
{
    CRIT_STATE_T rState;

    rState = x_crit_start();
#ifdef CC_MTAL_CB_REMOVE_LIST_FIXED
    prCbEv->u4RefMask &= ~(u4DoneMask);
    DBG_Printf(CB_DBG_GET_EVENT_FLG, "%s:pid %d/%d prCbEv %x, u4RefMask %d u4BitMask %d u4DoneMask %d fctid %d\n", __FUNCTION__,
               current->tgid, current->pid, prCbEv, prCbEv->u4RefMask, prCbEv->u4BitMask, u4DoneMask, prCbEv->rGetCb.eFctId);
#else
    prCbEv->u4BitMask &= ~(u4DoneMask);
#endif

    if ((prCbEv->u4BitMask == 0)
#ifdef CC_MTAL_CB_REMOVE_LIST_FIXED
    && (prCbEv->u4RefMask == 0)
#endif
        )
    {
        //if(prCbEv->rGetCb.eFctId == CB_MTAT_CEC_5V_STATUS_NFY)
        //{
        if(iCbCount > 0)
        {
            iCbCount--;
        }
        //}
        DLIST_REMOVE(prCbEv, &_qCbEvent, rLink);
        kfree(prCbEv);
#ifdef CC_MTAL_CB_REMOVE_LIST_FIXED
        DBG_Printf(CB_DBG_GET_EVENT_FLG, "%s:kfree %x\n", __FUNCTION__, prCbEv);
#endif
    }
    x_crit_end(rState);
}
#else
//-----------------------------------------------------------------------------
/** Get the first event from event queue.
 *  @return If success and has event in queue, pointer of the callback event.\n
 *          Otherwise, NULL.
 */
//-----------------------------------------------------------------------------
CB_CALLBACK_EVENT_T *_CB_GetEvent(void)
{
    CB_CALLBACK_EVENT_T *prLast=NULL;
    CRIT_STATE_T rState;
    BOOL fgFound=FALSE;
    static UINT32 cnt_o=0, cnt_sc=0;
    UINT32 i=0, j=0;

    rState = x_crit_start();
    DLIST_FOR_EACH(prLast, &_qCbEvent, rLink)
    {
        if (!prLast)
        {
            break;
        }

        if (prLast->rGetCb.eFctId >= CB_FCT_NUM)
        {
            break;
        }

        if (current->tgid == _arCbId[prLast->rGetCb.eFctId])
        {
            fgFound = TRUE;
            break;
        }
        i++;
        if (prLast->rGetCb.eFctId == CB_SC_TRIGGER)
        {
            j++;
        }

        if (prLast->rLink.pt_next == NULL)
        {
            break;
        }
    }
    if (fgFound && (prLast != NULL))
    {
        DLIST_REMOVE(prLast, &_qCbEvent, rLink);
    }
    cnt_o += i;
    cnt_sc += j;
    x_crit_end(rState);

    if (fgFound)
    {
        return prLast;
    }
    else
    {
        if (i == 0)
        {
            cnt_o = 0;
        }

        if (cnt_o > CB_OTHER_OVERFLOW)
        {
            if (_CB_doCbRegRecord())
            {
                printk("[Error] cb events of others overflow\n");
            }
        }

        if (j == 0)
        {
            cnt_sc = 0;
        }

        if (cnt_sc > SC_OTHER_OVERFLOW)
        {
            if (_CB_doCbRegRecord())
            {
                printk("[Error] cb events of sc overflow\n");
            }
        }

        return NULL;
    }
}

void _CB_FreeEvent(CB_CALLBACK_EVENT_T *prCbEv)
{
    if (prCbEv)
    {
        kfree(prCbEv);

        if (iCbCount > 0)
        {
            iCbCount--;
        }
    }

    return ;
}
#endif
//-----------------------------------------------------------------------------
/** Put the current thread into wait queue.
 */
//-----------------------------------------------------------------------------
#if 0
void _CB_PutThdIntoWQ(void)
{
    DECLARE_WAITQUEUE(wait, current);

    add_wait_queue(&_rWq, &wait);
    set_current_state(TASK_INTERRUPTIBLE);
    schedule();
    set_current_state(TASK_RUNNING);
    remove_wait_queue(&_rWq, &wait);
}
#endif

#ifdef CC_MTAL_MULTIPROCESS_CALLBACK
//-----------------------------------------------------------------------------
/** Clear specified PID entries in event list
 */
//-----------------------------------------------------------------------------
void _CB_ClearPendingEventEntries(void)
{
    CB_CALLBACK_EVENT_T *prLast=NULL;
    CRIT_STATE_T rState;
    BOOL fgFound;
    UINT32 u4DoneMask;

    do {
        fgFound = FALSE;

        rState = x_crit_start();
        if ((prLast = _CB_GetEvent(&u4DoneMask)) != NULL)
        {
            fgFound = TRUE;
#ifdef CC_MTAL_CB_REMOVE_LIST_FIXED
            DBG_Printf(CB_DBG_GET_EVENT_FLG, "_CB_ClearPendingFctEvent\n");
#endif
            _CB_FreeEvent(prLast, u4DoneMask);
        }
        x_crit_end(rState);
    } while (fgFound);
}

//-----------------------------------------------------------------------------
/** Clear entries with specified eFctId and PID in event list
 */
//-----------------------------------------------------------------------------
void _CB_ClearPendingFctEventEntries(CB_FCT_ID_T eFctId)
{
    CB_CALLBACK_EVENT_T *prLast=NULL;
    CRIT_STATE_T rState;
    BOOL fgFound;
    UINT32 u4DoneMask;

    do {
        fgFound = FALSE;

        rState = x_crit_start();
        DLIST_FOR_EACH(prLast, &_qCbEvent, rLink)
        {
            if (!prLast)
            {
                break;
            }

            if (prLast->rGetCb.eFctId >= CB_FCT_NUM)
            {
                break;
            }

            if (prLast->rGetCb.eFctId != eFctId)
            {
                continue;
            }

            if (current->tgid == _arCbId[prLast->rGetCb.eFctId])
            {
                u4DoneMask = BITMASK_SINGLE;
                fgFound = TRUE;
                break;
            }
            else if (current->tgid == _arMultiCbId[prLast->rGetCb.eFctId][0])
                {
                u4DoneMask = BITMASK_MULTI1;
                    fgFound = TRUE;
                    break;
                }
            else if (current->tgid == _arMultiCbId[prLast->rGetCb.eFctId][1])
            {
                u4DoneMask = BITMASK_MULTI2;
                fgFound = TRUE;
                break;
            }
            else if (current->tgid == _arMultiCbId[prLast->rGetCb.eFctId][2])
            {
                u4DoneMask = BITMASK_MULTI3;
                fgFound = TRUE;
                break;
            }
        }
        if (fgFound && (prLast != NULL))
        {
#ifdef CC_MTAL_CB_REMOVE_LIST_FIXED
            prLast->u4BitMask &= ~(u4DoneMask);
            DBG_Printf(CB_DBG_GET_EVENT_FLG, "_CB_ClearPendingFctEvent\n");
#endif
            _CB_FreeEvent(prLast, u4DoneMask);
        }
        x_crit_end(rState);
    } while (fgFound);
}
#else
//-----------------------------------------------------------------------------
/** Clear specified PID entries in event list
 */
//-----------------------------------------------------------------------------
void _CB_ClearPendingEventEntries(void)
{
    CB_CALLBACK_EVENT_T *prLast=NULL;
    CRIT_STATE_T rState;
    BOOL fgFound;

    do {
        fgFound = FALSE;

        rState = x_crit_start();
        DLIST_FOR_EACH(prLast, &_qCbEvent, rLink)
        {
            if (!prLast)
            {
                break;
            }

            if (prLast->rGetCb.eFctId >= CB_FCT_NUM)
            {
                break;
            }

            if (current->tgid == _arCbId[prLast->rGetCb.eFctId])
            {
                fgFound = TRUE;
                break;
            }

            if (prLast->rLink.pt_next == NULL)
            {
                break;
            }
        }
        if (fgFound && (prLast != NULL))
        {
            DLIST_REMOVE(prLast, &_qCbEvent, rLink);
            kfree(prLast);
        }
        x_crit_end(rState);
    } while (fgFound);
}

//-----------------------------------------------------------------------------
/** Clear entries with specified eFctId and PID in event list
 */
//-----------------------------------------------------------------------------
void _CB_ClearPendingFctEventEntries(CB_FCT_ID_T eFctId)
{
    CB_CALLBACK_EVENT_T *prLast=NULL;
    CRIT_STATE_T rState;
    BOOL fgFound;

    do {
        fgFound = FALSE;

        rState = x_crit_start();
        DLIST_FOR_EACH(prLast, &_qCbEvent, rLink)
        {
            if (!prLast)
            {
                break;
            }

            if (prLast->rGetCb.eFctId >= CB_FCT_NUM)
            {
                break;
            }

            if (current->tgid == _arCbId[prLast->rGetCb.eFctId])
            {
                if (prLast->rGetCb.eFctId == eFctId)
                {
                    fgFound = TRUE;
                    break;
                }
            }

            if (prLast->rLink.pt_next == NULL)
            {
                break;
            }
        }
        if (fgFound && (prLast != NULL))
        {
            DLIST_REMOVE(prLast, &_qCbEvent, rLink);
            kfree(prLast);
        }
        x_crit_end(rState);
    } while (fgFound);
}
#endif

//-----------------------------------------------------------------------------
/** Clear specified PID entries in CB PID table
 */
//-----------------------------------------------------------------------------
void _CB_ClearCbIdArray(void)
{
    INT32 i;
    CRIT_STATE_T rState;

    rState = x_crit_start();
    for (i=0; i<CB_FCT_NUM; i++)
    {
        if (current->tgid == _arCbId[i])
        {
            _CB_DbgReleaseRec((INT32)(current->tgid),(INT32)i);
            _arCbId[i] = (pid_t)0;
        }

        #ifdef CC_MTAL_MULTIPROCESS_CALLBACK
        if (current->tgid == _arMultiCbId[i][0])
        {
            _CB_DbgReleaseRec((INT32)(current->tgid),(INT32)i);
            _arMultiCbId[i][0] = (pid_t)0;
        }

        if (current->tgid == _arMultiCbId[i][1])
        {
            _CB_DbgReleaseRec((INT32)(current->tgid),(INT32)i);
            _arMultiCbId[i][1] = (pid_t)0;
        }

        if (current->tgid == _arMultiCbId[i][2])
        {
            _CB_DbgReleaseRec((INT32)(current->tgid),(INT32)i);
            _arMultiCbId[i][2] = (pid_t)0;
        }
        #endif
    }
    x_crit_end(rState);
}

