/* SPDX-License-Identifier: LGPL-2.1-only OR BSD-3-Clause */
/******************************************************************************
 *
 * This file is provided under a dual license.  When you use or
 * distribute this software, you may choose to be licensed under
 * version 2.1 of the GNU Lesser General Public License
 * ("LGPLv2.1 License") or BSD License.
 *
 * LGPLv2.1 License
 *
 * Copyright(C) 2019 MediaTek Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2.1 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See https://www.gnu.org/licenses/old-licenses/lgpl-2.1.html for more
 * details.
 *
 * BSD LICENSE
 *
 * Copyright(C) 2019 MediaTek Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *  * Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *****************************************************************************/

/*-----------------------------------------------------------------------------
 * $Author: dtvbm11 $
 * $Date: 2015/07/28 $
 * $RCSfile: cb_mpv.c,v $
 * $Revision: #1 $
 *---------------------------------------------------------------------------*/

/** @file cb_mpv.c
 *  This file implements callback interface of mpv for resolution change.
 */


//-----------------------------------------------------------------------------
// Include files
//-----------------------------------------------------------------------------

#include "cb_low.h"
#include "vdp_if.h"
#include <linux/module.h>


//-----------------------------------------------------------------------------
// Static functions
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/** MPV callback function.
 *  @param  pvNfyTag    Current EsId.
 *  @param  eNfyCond    flag: MPV_COND_RESOLUTION_CHG.
 *  @param  u4Data1     unused here.
 *  @param  u4Data2     unused here.
 */
//-----------------------------------------------------------------------------
static void _MPVCbFct(UINT8 u1VideoIndex,
	VDP_SOURCE_INFO_T rInfo)
{
    MPV_CB_T rMPVCb;        

    if(rInfo.u4SrcHeight > 1080)
    {
        rInfo.u4SrcHeight = 1080;
    }
    rMPVCb.u4EsId = 0;
    rMPVCb.u2W = rInfo.u4SrcWidth;
    rMPVCb.u2H = rInfo.u4SrcHeight;
    rMPVCb.ucProgressive = !(rInfo.u4Interlace);
    _CB_PutEvent(CB_MPV_TRIGGER, sizeof(MPV_CB_T), &rMPVCb);

}

//-----------------------------------------------------------------------------
/** To register/enable MPV callback function.
 *  @param  i4SubId     ID to identify subcomponent, or other parameter.
 *  @retval CBR_OK      Success.
 */
//-----------------------------------------------------------------------------

INT32 _MPV_RegTrigger(INT32 i4SubId)
{
    VDP_RegCbFunc(VDP_CB_FUNC_SRC_CHG_IND, (UINT32)_MPVCbFct, 0, 0);

    return CBR_OK;
}

