/* SPDX-License-Identifier: LGPL-2.1-only OR BSD-3-Clause */
/******************************************************************************
 *
 * This file is provided under a dual license.  When you use or
 * distribute this software, you may choose to be licensed under
 * version 2.1 of the GNU Lesser General Public License
 * ("LGPLv2.1 License") or BSD License.
 *
 * LGPLv2.1 License
 *
 * Copyright(C) 2019 MediaTek Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2.1 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See https://www.gnu.org/licenses/old-licenses/lgpl-2.1.html for more
 * details.
 *
 * BSD LICENSE
 *
 * Copyright(C) 2019 MediaTek Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *  * Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *****************************************************************************/

/*-----------------------------------------------------------------------------
 * $Author: brianpc.huang $
 * $Date: 2015/05/29 $
 * $RCSfile: cb_tim.c,v $
 * $Revision: #1 $
 *---------------------------------------------------------------------------*/

/** @file cb_tim.c
 *  This file implements callback interface of timer.
 */


//-----------------------------------------------------------------------------
// Include files
//-----------------------------------------------------------------------------

#include "cb_tim.h"
#include "cb_low.h"
#include <linux/module.h>
#include <linux/timer.h>


//-----------------------------------------------------------------------------
// Static variables
//-----------------------------------------------------------------------------

/// Timer list.
static struct timer_list timer;

/// Timer interval (ms).
static unsigned int _ui4Delay = 10;

/// Counter.
static unsigned int _ui4Counter = 0;


//-----------------------------------------------------------------------------
// Static functions
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/** Timer callback function.
 *  @param  param   Callback parameter, unused here.
 */
//-----------------------------------------------------------------------------
static void _TimerCbFct(unsigned long param)
{
    TIM_CB_T rTimCb;

    ++ _ui4Counter;
    timer.expires = jiffies + _ui4Delay * HZ / 1000;
    add_timer(&timer);

    rTimCb.i4Cnt = _ui4Counter;
    _CB_PutEvent(CB_TIM_TRIGGER, sizeof(TIM_CB_T), &rTimCb);
}

//-----------------------------------------------------------------------------
/** Initialize timer.
 */
//-----------------------------------------------------------------------------
static void _TimerInit(void)
{
    init_timer(&timer);

    timer.expires = jiffies + _ui4Delay * HZ / 1000;
    timer.function = _TimerCbFct;
    timer.data = 0;

    add_timer(&timer);
}


//-----------------------------------------------------------------------------
// Inter-file functions
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/** To register/enable timer callback function.
 *  @param  i4SubId     ID to identify subcomponent, or other parameter.
 *  @retval CBR_OK      Success.
 */
//-----------------------------------------------------------------------------
INT32 _TIM_RegTrigger(INT32 i4SubId)
{
    _ui4Delay = (UINT32)i4SubId;

    _TimerInit();

    return CBR_OK;
}

