/* SPDX-License-Identifier: LGPL-2.1-only OR BSD-3-Clause */
/******************************************************************************
 *
 * This file is provided under a dual license.  When you use or
 * distribute this software, you may choose to be licensed under
 * version 2.1 of the GNU Lesser General Public License
 * ("LGPLv2.1 License") or BSD License.
 *
 * LGPLv2.1 License
 *
 * Copyright(C) 2019 MediaTek Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2.1 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See https://www.gnu.org/licenses/old-licenses/lgpl-2.1.html for more
 * details.
 *
 * BSD LICENSE
 *
 * Copyright(C) 2019 MediaTek Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *  * Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *****************************************************************************/



#ifndef _KMEM_H_
#define _KMEM_H_


#include <linux/ioctl.h>
#include "u_handle2.h"

//fix89 #define DEV_KMEM        "/dev/kmem2"

#define IOCTL_KMEM_SET_MMAP_HANDLES     _IOW('k', 0, void *)
#define IOCTL_KMEM_GET_SIZE             _IOR('k', 1, size_t)
#define IOCTL_KMEM_GET_PAGE_OFFSET      _IOR('k', 2, size_t)
#define IOCTL_KMEM_ALIGNED_FREE         _IOW('k', 3, HANDLE_T)
#define IOCTL_KMEM_ALIGNED_ALLOC        _IOWR('k', 4, KMEM_ALIGNED_ALLOC_T)
#define IOCTL_KMEM_SLICED_ALIGNED_ALLOC _IOWR('k', 5, KMEM_SLICED_ALIGNED_ALLOC_T)
#define IOCTL_KMEM_GET_MMAP_HANDLES     _IOR('k', 6, void *)
/* add for map phy_addr to get chip info */
#define IOCTL_KMEM_GET_CHIP_INFO        _IOR('k', 7, void *)
/* add end for map phy_addr to get chip info */


typedef struct
{
    int n;
    HANDLE_T *handles;
} KMEM_MMAP_INFO_T;

typedef union
{
    struct
    {
        SIZE_T z_size;
        UINT32 u4Align;
    } in;
    struct
    {
        HANDLE_T handle;
    } out;
} KMEM_ALIGNED_ALLOC_T;

/* add for get chip info */
typedef struct
{
    HANDLE_T handle;
} KMEM_HANDLE_T;
/* add end for get chip info */


typedef union
{
    struct
    {
        UINT32 z_size;
        SIZE_T z_size_element;
        UINT32 u4Align;
    } in;
    struct
    {
        HANDLE_T handle;
    } out;
} KMEM_SLICED_ALIGNED_ALLOC_T;


#endif /* _KMEM_H_ */

