/* SPDX-License-Identifier: LGPL-2.1-only OR BSD-3-Clause */
/******************************************************************************
 *
 * This file is provided under a dual license.  When you use or
 * distribute this software, you may choose to be licensed under
 * version 2.1 of the GNU Lesser General Public License
 * ("LGPLv2.1 License") or BSD License.
 *
 * LGPLv2.1 License
 *
 * Copyright(C) 2019 MediaTek Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2.1 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See https://www.gnu.org/licenses/old-licenses/lgpl-2.1.html for more
 * details.
 *
 * BSD LICENSE
 *
 * Copyright(C) 2019 MediaTek Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *  * Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *****************************************************************************/

/*-----------------------------------------------------------------------------
 * $Author: brianpc.huang $
 * $Date: 2015/05/29 $
 * $RCSfile: mt53xx_gpio.h,v $
 * $Revision: #1 $
 *---------------------------------------------------------------------------*/

#ifndef MT53XX_GPIO_H
#define MT53XX_GPIO_H

/** @file mt53xx_gpio.h
 *  GPIO module for accessing GPIO pins
 */

//-----------------------------------------------------------------------------
// Include files
//-----------------------------------------------------------------------------

#include "x_typedef.h"
#include <linux/ioctl.h>


//-----------------------------------------------------------------------------
// Type definitions
//-----------------------------------------------------------------------------

/// GPIO_IOCTL_2ARG_T: Send 2 Arguments between user/kernel mode (ioctl interface)
typedef struct
{
    INT32 au4Arg[2];
} GPIO_IOCTL_2ARG_T;


//-----------------------------------------------------------------------------
// Configurations
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// Constant definitions
//-----------------------------------------------------------------------------

// GPIO IOCTYPE
#define GPIO_IOCTYPE                    0

// GPIO commands
#define GPIO_CMD_INIT                   0
#define GPIO_CMD_STOP                   1
#define GPIO_CMD_SETOUT                 2
#define GPIO_CMD_GETIN                  3

// GPIO IOCTL numbers
#define GPIO_IOCTL_INIT                 _IO(GPIO_IOCTYPE, GPIO_CMD_INIT)
#define GPIO_IOCTL_STOP                 _IO(GPIO_IOCTYPE, GPIO_CMD_STOP)
#define GPIO_IOCTL_SETOUT               _IOW(GPIO_IOCTYPE, GPIO_CMD_SETOUT, GPIO_IOCTL_2ARG_T)
#define GPIO_IOCTL_GETIN                _IOW(GPIO_IOCTYPE, GPIO_CMD_GETIN, INT32)


//-----------------------------------------------------------------------------
// Prototype  of public functions
//-----------------------------------------------------------------------------

extern INT32 GPIO_Init(void);
extern INT32 GPIO_Stop(void);
extern INT32 GPIO_SetOut(INT32 i4GpioNum, const INT32 i4Value);
extern INT32 GPIO_GetIn(INT32 i4GpioNum);


#endif /* MT53XX_GPIO_H */
