/* SPDX-License-Identifier: LGPL-2.1-only OR BSD-3-Clause */
/******************************************************************************
 *
 * This file is provided under a dual license.  When you use or
 * distribute this software, you may choose to be licensed under
 * version 2.1 of the GNU Lesser General Public License
 * ("LGPLv2.1 License") or BSD License.
 *
 * LGPLv2.1 License
 *
 * Copyright(C) 2019 MediaTek Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2.1 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See https://www.gnu.org/licenses/old-licenses/lgpl-2.1.html for more
 * details.
 *
 * BSD LICENSE
 *
 * Copyright(C) 2019 MediaTek Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *  * Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *****************************************************************************/


/** @file mt53xx_i1394.h
 *  Define I1394 interface.
 */

#ifndef MT53XX_I1394_H
#define MT53XX_I1394_H

//-----------------------------------------------------------------------------
// Include files
//-----------------------------------------------------------------------------

#include <x_typedef.h>


//-----------------------------------------------------------------------------
// Configurations
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// Constant definitions
//-----------------------------------------------------------------------------

#define I1394_DEVICE                            "/dev/dif"

#define STATUS_SUCCESS                          0
#define STATUS_INSUFFICIENT_RESOURCES           -1

#define I1394_IOCTL_CTRL_GENERAL                0
#define I1394_IOCTL_CTRL_PLAY                   1
#define I1394_IOCTL_CTRL_STOP                   2

//-----------------------------------------------------------------------------
// Type definitions
//-----------------------------------------------------------------------------

typedef struct _I1394_REMOTE_DEV
{
    INT32 i4NodeID;
    INT32 i4Type;
    UCHAR uDevName[64];
} I1394_REMOTE_DEV, *PI1394_REMOTE_DEV;

// Structure of 1394 device list
typedef struct
{
    INT32  i4TotalDevice;
    INT32  i4RetValue;
} DEVICE_INFO_HEADER;

typedef struct
{
    DEVICE_INFO_HEADER rInfoHeader;
    PI1394_REMOTE_DEV  prDeviceList;
} I1394_DEVICE_LIST_INFO_T;

// Structure of get all device command
typedef struct
{
    UINT32  u4BufferSize;
    void*   pTransferBuf;
} I1394_IOCTL_ALL_DEVICE_GET_T;

// Structure of get device number command
typedef struct
{
    UINT32 *pu4DeviceNo;
} I1394_IOCTL_DEVICE_CNT_GET_T;

// Structure of set device format command
typedef struct
{
    UINT32 u4NodeId;
    BOOL*  pbIsDVMode;
} I1394_IOCTRL_DEVICE_FMT_SET_T;

// Structure of set device control command
typedef struct
{
    UINT32 u4NodeId;
    UINT32 u4CtrlCode;
    UINT32 u4Param1;
    UINT32 u4Param2;
} I1394_IOCTRL_DEVICE_CONTROL_T;

// Structure of set iso. channel control command
typedef struct
{
    BOOL bIsRxTransaction;
    BOOL bIsTransactionEnable;
} I1394_IOCTRL_ISO_CONTROL_T;

// Structure of get device play state command
typedef struct
{
    UINT32 u4NodeId;
    BOOL   bIsDVMode;
    INT32  *pi4DeviceState;
} I1394_IOCTRL_DEVICE_STATE_T;


//-----------------------------------------------------------------------------
// Macro definitions
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// Prototypes of public functions
//-----------------------------------------------------------------------------


#endif  // MT53XX_I1394_H

