/* SPDX-License-Identifier: LGPL-2.1-only OR BSD-3-Clause */
/******************************************************************************
 *
 * This file is provided under a dual license.  When you use or
 * distribute this software, you may choose to be licensed under
 * version 2.1 of the GNU Lesser General Public License
 * ("LGPLv2.1 License") or BSD License.
 *
 * LGPLv2.1 License
 *
 * Copyright(C) 2019 MediaTek Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2.1 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See https://www.gnu.org/licenses/old-licenses/lgpl-2.1.html for more
 * details.
 *
 * BSD LICENSE
 *
 * Copyright(C) 2019 MediaTek Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *  * Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *****************************************************************************/


#ifndef _X_KMEM_H_
#define _X_KMEM_H_


/*-----------------------------------------------------------------------------
                    include files
-----------------------------------------------------------------------------*/

#ifdef __KERNEL__
#include <linux/compiler.h>
#endif /* __KERNEL__ */
#include "u_handle2.h"

/*-----------------------------------------------------------------------------
                    functions declarations
 ----------------------------------------------------------------------------*/

#ifdef __cplusplus
extern "C" {
#endif

#ifndef __KERNEL__

extern VOID *x_kmem_alloc(SIZE_T z_size);
extern VOID *x_kmem_calloc(UINT32 ui4_num_element, SIZE_T z_size_element);
extern VOID *x_kmem_realloc(VOID *pv_mem_block, SIZE_T z_new_size);
extern VOID x_kmem_free(VOID *pv_mem_block);
extern VOID *x_kmem_aligned_alloc(SIZE_T z_size, UINT32 u4Align);
extern VOID x_kmem_aligned_free(VOID *pv_mem);

#endif /* __KERNEL__ */

extern HANDLE_T x_kmem_block_alloc(SIZE_T z_size);
extern HANDLE_T x_kmem_block_sliced_alloc(UINT32 z_size, SIZE_T z_size_element);
extern HANDLE_T x_kmem_block_sliced_aligned_alloc(UINT32 z_size, SIZE_T z_size_element, UINT32 u4Align);
extern VOID x_kmem_block_free(HANDLE_T handle);
extern HANDLE_T x_kmem_block_aligned_alloc(SIZE_T z_size, UINT32 u4Align);
extern VOID x_kmem_block_aligned_free(HANDLE_T handle);
extern VOID *x_kmem_block_map(HANDLE_T handle);
extern VOID *x_kmem_block_map_fd(int fd, HANDLE_T handle);
extern VOID *x_kmem_block_map_ext(void *handle);
extern VOID *x_kmem_block_multimap(int n, HANDLE_T handles[]);
extern VOID x_kmem_block_unmap(VOID *pv_mem_block);
extern VOID x_kmem_block_unmap_fd(int fd, VOID *pv_mem_block);
extern VOID x_kmem_block_unmap_ext(VOID *addr, size_t size);
/* add for get chip info */
extern VOID *x_kmem_map_chip_info_addr(VOID);
extern VOID x_kmem_unmap_chip_info_addr(VOID *pv_vir_addr);
/* add end for get chip info */




extern SIZE_T x_kmem_calc_slice_size(UINT32 z_size);

#define x_kmem_block_auto_sliced_alloc(z_size) \
    x_kmem_block_sliced_alloc(z_size, x_kmem_calc_slice_size(z_size))
#define x_kmem_block_auto_sliced_aligned_alloc(z_size, u4Align) \
    x_kmem_block_sliced_aligned_alloc(z_size, x_kmem_calc_slice_size(z_size), u4Align)

#ifdef __KERNEL__

typedef struct
{
    int n;
    size_t size, last_size;
} KMEM_BLK_INFO;

extern INT32 x_kmem_block_info(HANDLE_T handle, KMEM_BLK_INFO *kmem_blk_info);
extern HANDLE_T x_kmem_block_handle(VOID *pv_mem);
extern VOID *x_kmem_addr_user_to_kernel(VOID __user *pv_user);
extern VOID *x_kmem_kernel_address(HANDLE_T handle, UINT32 ui4_ofst);
extern HANDLE_T x_kmem_create_handle(VOID *pv_mem, SIZE_T z_size);
extern VOID x_kmem_destroy_handle(HANDLE_T handle);
extern VOID *x_mem_aligned_alloc(SIZE_T z_size, UINT32 u4Align);
extern VOID x_mem_aligned_free(VOID *pv_mem);
extern VOID *x_mem_valloc(UINT32 u4Size);
extern VOID x_mem_vfree(VOID *pv_mem);

#endif /* __KERNEL__ */

#ifdef __cplusplus
}
#endif

#endif /* _X_KMEM_H_ */
