/* SPDX-License-Identifier: LGPL-2.1-only OR BSD-3-Clause */
/******************************************************************************
 *
 * This file is provided under a dual license.  When you use or
 * distribute this software, you may choose to be licensed under
 * version 2.1 of the GNU Lesser General Public License
 * ("LGPLv2.1 License") or BSD License.
 *
 * LGPLv2.1 License
 *
 * Copyright(C) 2019 MediaTek Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2.1 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See https://www.gnu.org/licenses/old-licenses/lgpl-2.1.html for more
 * details.
 *
 * BSD LICENSE
 *
 * Copyright(C) 2019 MediaTek Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *  * Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *****************************************************************************/

/*-----------------------------------------------------------------------------
 * $Author: dtvbm11 $
 * $Date: 2015/07/28 $
 * $RCSfile: fops_gpio.c,v $
 * $Revision: #1 $
 *---------------------------------------------------------------------------*/


/** @file gpio_mod.c
 *  Kernel mode GPIO module driver to handle ioctl from user sapce.
 */

 
//-----------------------------------------------------------------------------
// Include files
//-----------------------------------------------------------------------------
/*lint -save -e* */
#include "ionodes.h"
#ifdef CONFIG_64BIT
#include <linux/soc/mediatek/x_linux.h>
#else
#include "x_linux.h"
#endif
#include "x_assert.h"
#include <linux/version.h>
#include <linux/mt53xx_gpio.h>
#include <linux/semaphore.h>
/*lint -restore */

//-----------------------------------------------------------------------------
// Configurations
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// Constant definitions
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// Type definitions
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// Macro definitions
//-----------------------------------------------------------------------------

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,37)
#define DECLARE_MUTEX(mutex)  DEFINE_SEMAPHORE(mutex)
#endif

//-----------------------------------------------------------------------------
// Imported variables
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// Imported functions
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// Static function forward declarations
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// Static variables
//-----------------------------------------------------------------------------
/*lint -save -e* */
static DECLARE_MUTEX(_rMutex);
/*lint -restore */


//-----------------------------------------------------------------------------
// Static functions
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// Inter-file functions
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// Public functions
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
/** gpio_ioctl() GPIO driver ioctl function.
 */
//-----------------------------------------------------------------------------
/*lint -save -e* */
long gpio_ioctl(struct file *file, unsigned int cmd,
                unsigned long arg)
/*lint -restore */
{
    GPIO_IOCTL_2ARG_T rArg;
    INT32 i4Ret;

    if (down_interruptible(&_rMutex))
    {
        return (-ERESTARTSYS);
    }
    
    switch (cmd) {
    case GPIO_IOCTL_INIT:
        i4Ret = GPIO_Init();
        break;
    case GPIO_IOCTL_STOP:
        i4Ret = GPIO_Stop();
        break;
    case GPIO_IOCTL_SETOUT:
        if (!access_ok(VERIFY_READ, (void __user *)arg,
                       sizeof(GPIO_IOCTL_2ARG_T)))
        {
            i4Ret = -EFAULT;
            goto _GPIO_IOCTL_EXIT;
        }
        if (copy_from_user(&rArg, (void __user *)arg,
                           sizeof(GPIO_IOCTL_2ARG_T)))
        {
            i4Ret = -EFAULT;
            goto _GPIO_IOCTL_EXIT;
        }
        i4Ret = GPIO_SetOut(rArg.au4Arg[0], rArg.au4Arg[1]);
        break;
    case GPIO_IOCTL_GETIN:
        if (!access_ok(VERIFY_READ, (void __user *)arg,
                       sizeof(INT32)))
        {
            i4Ret = -EFAULT;
            goto _GPIO_IOCTL_EXIT;
        }
        if (copy_from_user(&rArg.au4Arg[0], (void __user *)arg,
                           sizeof(INT32)))
        {
            i4Ret = -EFAULT;
            goto _GPIO_IOCTL_EXIT;
        }
        i4Ret = GPIO_GetIn(rArg.au4Arg[0]);
        break;
    default:
        i4Ret = -EINVAL;
        ASSERT(0);
    }

_GPIO_IOCTL_EXIT:

    up(&_rMutex);
    
    return i4Ret;
}

struct file_operations gpio_fops = {
    .unlocked_ioctl      = gpio_ioctl,
};
