/* SPDX-License-Identifier: LGPL-2.1-only OR BSD-3-Clause */
/******************************************************************************
 *
 * This file is provided under a dual license.  When you use or
 * distribute this software, you may choose to be licensed under
 * version 2.1 of the GNU Lesser General Public License
 * ("LGPLv2.1 License") or BSD License.
 *
 * LGPLv2.1 License
 *
 * Copyright(C) 2019 MediaTek Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2.1 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See https://www.gnu.org/licenses/old-licenses/lgpl-2.1.html for more
 * details.
 *
 * BSD LICENSE
 *
 * Copyright(C) 2019 MediaTek Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *  * Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *****************************************************************************/

/*-----------------------------------------------------------------------------
 *
 * $Author: dtvbm11 $
 * $Date: 2015/07/28 $
 * $RCSfile: $
 * $Revision: #1 $
 *
 *---------------------------------------------------------------------------*/

/** @file fops_tsrec.c
 *  Module interface of TS record driver
 */


//-----------------------------------------------------------------------------
// Include files
//-----------------------------------------------------------------------------

#include "ionodes.h"
#include "x_linux.h"
#include "fbm_drvif.h"
#include <linux/version.h>

//-----------------------------------------------------------------------------
// Static function forward declarations
//-----------------------------------------------------------------------------

static int tsrec_mmap(struct file *, struct vm_area_struct *);

//---------------------------------------------------------------------------
// Public functions
//---------------------------------------------------------------------------
//-----------------------------------------------------------------------------
/** _TsRecOnMMAPAddr
 *
 *
 *  @param  arg             Command-specific argument
 *
 *  @retval TRUE	        Succeed
 *  @retval FALSE	        Fail
 */
//-----------------------------------------------------------------------------

UINT32 _TsRecOnMMAPAddr(void)
{
	FBM_POOL_T *prFbmPool;

    // Fixme, temp use PIP scaler
	//prFbmPool = FBM_GetPoolInfo((UINT8)FBM_POOL_TYPE_DMX_REC_TS);
	//prFbmPool = FBM_GetPoolInfo((UINT8)FBM_POOL_TYPE_SCPOS_PIP);
	//prFbmPool = FBM_GetPoolInfo((UINT8)FBM_POOL_TYPE_NR_1);
	prFbmPool = FBM_GetPoolInfo((UINT8)FBM_POOL_TYPE_PVR);

	return PAGE_ALIGN(prFbmPool->u4Addr);	//jpg driver internal used

}

//-----------------------------------------------------------------------------
/** _TsRecOnMMAPSize
 *
 *
 *  @param  arg             Command-specific argument
 *
 *  @retval TRUE	        Succeed
 *  @retval FALSE	        Fail
 */
//-----------------------------------------------------------------------------
UINT32 _TsRecOnMMAPSize(void)
{
	FBM_POOL_T *prFbmPool;

	//prepare input buffer from FBM, demo purpose
	//prFbmPool = FBM_GetPoolInfo((UINT8)FBM_POOL_TYPE_DMX_REC_TS);
	//prFbmPool = FBM_GetPoolInfo((UINT8)FBM_POOL_TYPE_SCPOS_PIP);
	//prFbmPool = FBM_GetPoolInfo((UINT8)FBM_POOL_TYPE_NR_1);
	prFbmPool = FBM_GetPoolInfo((UINT8)FBM_POOL_TYPE_PVR);

	return (prFbmPool->u4Size & PAGE_MASK);
}

//---------------------------------------------------------------------------
// Public functions
//---------------------------------------------------------------------------

static int tsrec_mmap(struct file *file, struct vm_area_struct *vma)
{
	unsigned long off, start;
	u32 len;

	if (vma->vm_pgoff > (~0UL >> PAGE_SHIFT))
		return -EINVAL;

	off = vma->vm_pgoff << PAGE_SHIFT;

	start = (unsigned long)_TsRecOnMMAPAddr(); //info->fix.smem_start;
	//info->fix.smem_len;
	len = PAGE_ALIGN(start & ~PAGE_MASK) + _TsRecOnMMAPSize();
	start &= PAGE_MASK;

	if ((vma->vm_end - vma->vm_start + (u64)off) > len)
	{
		return -EINVAL;
	}
	off += start;
	vma->vm_pgoff = off >> PAGE_SHIFT;

	/* This is an IO map - tell maydump to skip this VMA */
	vma->vm_flags |= VM_IO;

	vma->vm_page_prot = pgprot_writecombine(vma->vm_page_prot);

	/*
	 * Don't alter the page protection flags; we want to keep the area
	 * cached for better performance.  This does mean that we may miss
	 * some updates to the screen occasionally, but process switches
	 * should cause the caches and buffers to be flushed often enough.
	 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
	if (io_remap_pfn_range(vma, vma->vm_start, (off >> PAGE_SHIFT),
				vma->vm_end - vma->vm_start,
				vma->vm_page_prot))
#else
    if (io_remap_page_range(vma, vma->vm_start, off,
        vma->vm_end - vma->vm_start,
        vma->vm_page_prot))
#endif
	{
		return -EAGAIN;
	}

	return 0;
}

struct file_operations tsrec_fops = {
	.owner   = THIS_MODULE,
	.mmap  = tsrec_mmap,
};

