/* SPDX-License-Identifier: LGPL-2.1-only OR BSD-3-Clause */
/******************************************************************************
 *
 * This file is provided under a dual license.  When you use or
 * distribute this software, you may choose to be licensed under
 * version 2.1 of the GNU Lesser General Public License
 * ("LGPLv2.1 License") or BSD License.
 *
 * LGPLv2.1 License
 *
 * Copyright(C) 2019 MediaTek Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2.1 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See https://www.gnu.org/licenses/old-licenses/lgpl-2.1.html for more
 * details.
 *
 * BSD LICENSE
 *
 * Copyright(C) 2019 MediaTek Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *  * Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *****************************************************************************/

#ifdef CONFIG_64BIT
#include <linux/soc/mediatek/x_linux.h>
#else
#include "x_linux.h"
#endif
#include "driver_os.h"
#include "isr_linux_drv.h"
#include "x_assert.h"
#include "x_hal_5381.h"

extern INT32 ev_grp_init (VOID);
extern INT32 msg_q_init (VOID);
extern INT32 os_sema_init (VOID);
extern INT32 os_thread_init (VOID);
extern INT32 os_timer_init (VOID);

/*----------------------------------------------------------------------------
 * Function: x_rtos_init
 *
 * Description:
 *      Initialize OS driver. This function is called during system
 *      initialization.
 *
 * Inputs: -
 *
 * Outputs: -
 *
 * Returns:
 *      OSR_DRV_OK: Successful
 *      OSR_DRV_FAIL: Failed
 *---------------------------------------------------------------------------*/

static INT32 x_rtos_init(void)
{
    // Event group
    if (ev_grp_init() != OSR_OK)
    {
        return OSR_DRV_FAIL;
    }
   
    // Message queue 
    if (msg_q_init() != OSR_OK)
    {
        return OSR_DRV_FAIL;
    }

    // ISR library
    if (os_drv_isr_init() != OSR_DRV_OK)
    {
        return OSR_DRV_FAIL;
    }

    // Semaphore
    if (os_sema_init() != OSR_OK)
    {
        return OSR_DRV_FAIL;
    }

    // Thread
    if (os_thread_init() != OSR_OK)
    {
        return OSR_DRV_FAIL;
    }

    // Timer
    if (os_timer_init() != OSR_OK)
    {
        return OSR_DRV_FAIL;
    }

    return OSR_DRV_OK;
}

/*----------------------------------------------------------------------------
 * Function: x_rtos_exit
 *
 * Description:
 *      Shutdown OS driver. This function is called while system shutdown.
 *
 * Inputs: -
 *
 * Outputs: -
 *
 * Returns:
 *      TRUE: Successful
 *      FALSE: Failed
 *---------------------------------------------------------------------------*/
BOOL x_rtos_exit(void)
{
    BOOL fgRet = TRUE;

    return fgRet;
}

void OSDRV_Init(void)
{
    x_rtos_init();
}
