/* SPDX-License-Identifier: LGPL-2.1-only OR BSD-3-Clause */
/******************************************************************************
 *
 * This file is provided under a dual license.  When you use or
 * distribute this software, you may choose to be licensed under
 * version 2.1 of the GNU Lesser General Public License
 * ("LGPLv2.1 License") or BSD License.
 *
 * LGPLv2.1 License
 *
 * Copyright(C) 2019 MediaTek Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2.1 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See https://www.gnu.org/licenses/old-licenses/lgpl-2.1.html for more
 * details.
 *
 * BSD LICENSE
 *
 * Copyright(C) 2019 MediaTek Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *  * Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *****************************************************************************/

/*-----------------------------------------------------------------------------
 * Description:
 *---------------------------------------------------------------------------*/

/* driver-OS header files */
#include "x_bim.h"
#define __USE_XBIM_IRQS        // UGLY HACK! Make sure we don't include mach/irqs.h from kernel.

#include "x_linux.h"
#include "driver_os.h"
#include "isr_linux_drv.h"
#include "x_assert.h"
#include "x_printf.h"
#include "x_timeprofile.h"

#include <linux/version.h>
#include <linux/interrupt.h>

#ifndef SA_SHIRQ		/* depricated naming*/
#define SA_SHIRQ		IRQF_SHARED
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 4, 0)
#define SA_INTERRUPT		IRQF_DISABLED
#else
#define SA_INTERRUPT		0
#endif
#endif

//---------------------------------------------------------------------------

#if defined(CC_MT5396)||defined(CC_MT5368)||defined(CC_MT5389)
    #define MAX_VECTOR_ID       72                  // Max vector ID
#else
    #define MAX_VECTOR_ID       MAX_IRQ_VECTOR
#endif

#if defined(CC_MT5396)||defined(CC_MT5368)||defined(CC_MT5399)||defined(CC_MT5398)||defined(CC_MT5880)||defined(CC_MT5890)
    #define SPI_OFFSET          32
#elif defined(CC_MT5891)
	#define SPI_OFFSET          104
#elif defined(CC_MT5882)||defined(CC_MT5883)
    #define SPI_OFFSET          64
#else
    #define SPI_OFFSET          0
#endif

#define ISR_VECTOR_SIZE     (MAX_VECTOR_ID + 1) // Max number of registered vectors

typedef struct
{
    x_os_isr_fct    pf_isr;
    char            ucName[32];
} IRQ_ARRAY_T;

#if defined(CC_MT5396)||defined(CC_MT5368)||defined(CC_MT5389)||defined(CC_MT5399)||defined(CC_MT5398)||defined(CC_MT5880)||defined(CC_MT5860)||defined(CC_MT5881)||defined(CC_MT5890)||defined(CC_MT5882)||defined(CC_MT5883)||defined(CC_MT5891)
IRQ_ARRAY_T rIrqs[ISR_VECTOR_SIZE] =
{
    {NULL, "PDWNC"},
    {NULL, "SERIAL"},
    {NULL, "NAND"},
    {NULL, "T0"},
    {NULL, "T1"},
    {NULL, "T2"},
    {NULL, "SMARTCARD"},
    {NULL, "WDT"},
    {NULL, "ENET"},
    {NULL, "MJC"},
    {NULL, "PSCAN"},
    {NULL, "USB1"},
    {NULL, "IMGRZ"},
    {NULL, "VENC"},
    {NULL, "SPDIF_PSR"},
    {NULL, "USB0"},
    {NULL, "AUDIO"},
    {NULL, "RS232"},
    {NULL, "LED"},
    {NULL, "OSD"},
    {NULL, "VDOIN"},
    {NULL, "BLK2RS"},
    {NULL, "FLASH"},
    {NULL, "POST_PROC"},
    {NULL, "VDEC"},
    {NULL, "GFX"},
    {NULL, "DEMUX"},
    {NULL, "DEMOD"},
    {NULL, "FCI"},
    {NULL, "APWM"},
    {NULL, "MISC2"},
    {NULL, "MISC"},
    //
    //  MISC_IRQ
    //
    {NULL, "DRAMC"},
    {NULL, "EXT1"},
    {NULL, "POD"},
    // MISC_IRQ[4:3]
    #if defined(CC_MT5891)
    {NULL, "PMU0"},
    {NULL, "PMU1"},
    #elif defined(CC_MT5399)||defined(CC_MT5398)||defined(CC_MT5880)||defined(CC_MT5881)||defined(CC_MT5890)||defined(CC_MT5882)||defined(CC_MT5389)||defined(CC_MT5396)
    {NULL, "PMU1"},
    {NULL, "PMU0"},
    #elif defined(CC_MT5883)
    {NULL, "EXT_INT_FIQ"},
    {NULL, "PMU1"},
    #else 
    #error PMU IRQ array not set.
    #endif
    // MISC_IRQ[5]
    {NULL, "DRAMC_CHB"},
    {NULL, "TVE"},
    // MISC_IRQ[7]
    #if defined(CC_MT5891)
    {NULL, "SCALER"},    
    #else
    {NULL, "PCIE"},
    #endif
    // MISC_IRQ[8]
    {NULL, "APWM2"},
    {NULL, "SPI"},
    {NULL, "JPGDEC"},
    {NULL, "P"},
    // MISC_IRQ[12]
    #if defined(CC_MT5882)||defined(CC_MT5883)||defined(CC_MT5891)
	{NULL, "GCPU"},
    #else
	{NULL, "AES"},
    #endif
    #if defined(CC_MT5399)||defined(CC_MT5890)||defined(CC_MT5882)||defined(CC_MT5883)||defined(CC_MT5891)
    {NULL, "TCON_EH"},
    {NULL, "TCON_TCH"},
    #else
    {NULL, "TCON_TCH"},
    {NULL, "TCON_EH"},
    #endif
    // MISC_IRQ[15]
    {NULL, "OD"},
    #if defined(CC_MT5882)||defined(CC_MT5883)||defined(CC_MT5891)
	{NULL, "LED_ADD"},
    #else
    {NULL, "USB2"},
    #endif
    // MISC_IRQ[17]
    {NULL, "USB3"},
    {NULL, "GDMA"},
    {NULL, "EPHY"},
    {NULL, "TCPIP"},
    // MISC_IRQ[21]
    #if defined(CC_MT5883)
    {NULL, "I2C"},
    #else
    {NULL, "GFX3D_GP"},
    #endif
    // MISC_IRQ[22]
    {NULL, "GFX3D_PP"},
    {NULL, "GFX3D_PMU"},
    // MISC_IRQ[24]
    #if defined(CC_MT5882)||defined(CC_MT5883)||defined(CC_MT5891)
	{NULL, "SSUSB"},
	{NULL, "R2R"},
    #else
    {NULL, "GFX3D_GPMMU"},
    {NULL, "GFX3D_PPMMU"},
    #endif
    // MISC_IRQ[26]
    {NULL, "MMU_GFX"},
    {NULL, "MMU_GCPU"},
    {NULL, "MMU_IMGRZ"},
    {NULL, "MMU_JPG"},
    {NULL, "MMU_GDMA"},
    {NULL, "DDI"},
    //
    //  MISC2_IRQ
    //        
    {NULL, "LZHS"},
    {NULL, "L2C:SFLASH"},
    {NULL, "IMGRZ2"},
    {NULL, "MMU_IMGRZ2"},
    {NULL, "MSDC2"},
    {NULL, "PNG1"},
    {NULL, "PNG2"},
    {NULL, "PNG3"},
    {NULL, "MMU_PNG1"},
    // MISC2_IRQ[9]
    #if defined(CC_MT5399)||defined(CC_MT5890)||defined(CC_MT5882)||defined(CC_MT5883)||defined(CC_MT5891)
    {NULL, "SMCARD1"},
    {NULL, "TDDC"},
    {NULL, "APWM3"},
    // MISC2_IRQ[12]
    #if defined(CC_MT5882)||defined(CC_MT5883)||defined(CC_MT5891)
	{NULL, "IRT_DMA"},
	#else
    {NULL, "MMU_VDEC:IRT_DMA"},
    #endif
    // MISC2_IRQ[13]
    {NULL, "MMU_DDI"},
    {NULL, "MMU_DMX"},
    {NULL, "VECTOR_USB3_1"},
    {NULL, "RS232_1"},
    // MISC2_IRQ[17]       
    #ifdef CC_MT5399
    {NULL, "APOLLO"},
    #else // for CC_MT5890
    {NULL, "MMU_IMGRZ3"},
    {NULL, "IMGRZ3"},
    // MISC2_IRQ[19]          
    #if defined(CC_MT5882)||defined(CC_MT5883)
	{NULL, "RS232_0"},
	{NULL, "OSD_INT2"},
    {NULL, "OSD_INT3"},
    {NULL, "NULL"},
	#else
    {NULL, "ZZ:RS232_0"},
    {NULL, "DRAMC_CHC:OSD_INT2:VENC_LT"},
    {NULL, "VDEC_LAE0:OSD_INT3"},
    {NULL, "VDEC_LAE1"},
    #endif
    // MISC2_IRQ[23]  
    {NULL, "TRNG"},
    {NULL, "P_B"},
    // MISC2_IRQ[25]          
    #if defined(CC_MT5882)||defined(CC_MT5883)
	{NULL, "VDEC_CORE1"},
	#else
    {NULL, "VDEC_CORE1:VDEC_IOMMU"},
    #endif
    // MISC2_IRQ[26]   
    {NULL, "ABIM0"},
    {NULL, "ABIM1"},
    {NULL, "ABIM2"},
    {NULL, "ABIM3"},
    {NULL, "CI_FMR"},
    {NULL, "CI_FMR2"},
    //
    //  MISC3_IRQ
    //    
    #if defined(CC_MT5890)
    // MISC3_IRQ[0]
    {NULL, "PTP_THERM:THERMAL0"},
    {NULL, "PTP_FSM:THERMAL1"},
    {NULL, "SC_TIMER"},
    {NULL, "HDMIRX_20"},
    {NULL, "NULL:G3D_IRQ_PP_0"},
    {NULL, "NULL:G3D_IRQ_PP_1"},
    {NULL, "NULL:G3D_IRQ_PP_2"},
    {NULL, "NULL:G3D_IRQ_PP_3"},
    {NULL, "NULL:G3D_IRQ_PP_ALL"},
    {NULL, "NULL:G3D_IRQ_GP"},
    {NULL, "NULL:G3D_IRQ_GPMMU"},
    {NULL, "NULL:G2D_IRQ_GPMMU_0"},
    {NULL, "NULL:G1D_IRQ_GPMMU_1"},
    {NULL, "NULL:G0D_IRQ_GPMMU_2"},
    {NULL, "NULL:G1D_IRQ_GPMMU_3"},
    #elif defined(CC_MT5882)||defined(CC_MT5883)
    // MISC3_IRQ[0]
	{NULL, "THERMAL0"},
    {NULL, "THERMAL1"},
    {NULL, "SC_TIMER"},
    {NULL, "NULL"},
    {NULL, "G3D_IRQ_PP_0"},
    {NULL, "G3D_IRQ_PP_1"},
    {NULL, "G3D_IRQ_PP_2"},
    {NULL, "G3D_IRQ_PP_3"},
    {NULL, "G3D_IRQ_PP_ALL"},
    {NULL, "G3D_IRQ_PP_GP"},
    {NULL, "G3D_IRQ_PP_GPMMU"},
    {NULL, "G2D_IRQ_PP_GPMMU_0"},
    {NULL, "G1D_IRQ_PP_GPMMU_1"},
    {NULL, "G0D_IRQ_PP_GPMMU_2"},
    {NULL, "G1D_IRQ_PP_GPMMU_3"},
    #elif defined(CC_MT5891)
    // MISC3_IRQ[0]    
	{NULL, "THERMAL0"},
    {NULL, "THERMAL1"},
    {NULL, "SC_TIMER"},
    {NULL, "DRAMC_C"},
    {NULL, "EXT_INT0"},
    {NULL, "EXT_INT1"},
    {NULL, "EXT_INT2"},
    {NULL, "OSD_VSYNC_INT"},
    // MISC3_IRQ[8]      
    {NULL, "OSD2_INT"},
    {NULL, "OSD3_INT"},
    {NULL, "UFOZIP_INT"},
    {NULL, "ecbus_int"},
    {NULL, "MMU_IMGRZ4"},
    {NULL, "IMGRZ4"},
    {NULL, "UFOZIP_MMU_INT"}, 
    {NULL, "arm_audio0_int"},   
    // MISC3_IRQ[16]         
    {NULL, "HDMIRX_A_20"},
    {NULL, "HDMIRX_B_20"},
    {NULL, "HDMIRX_C_20"},
    {NULL, "HDMIRX_D_20"},
    {NULL, "USB3_0d"},
    {NULL, "NULL"},
    {NULL, "arm_audio1_int"}, 
    {NULL, "arm_audio2_int"},       
    // MISC3_IRQ[24]         
    {NULL, "arm_audio3_int"},
    {NULL, "arm_audio4_int"},
    {NULL, "arm_audio5_int"},
    {NULL, "arm_audio6_int"},
    {NULL, "arm_audio7_int"},
    {NULL, "arm_audio8_int"},
    {NULL, "arm_audio9_int"}, 
    {NULL, "arm_audio10_int"},    
    // MISC5_IRQ[0]         
    {NULL, "aout_2risc_intr"},
    {NULL, "aout2_2risc_intr"},
    {NULL, "aout3_2risc_intr"},
    {NULL, "iec_2risc_intr"},
    {NULL, "asrc_2risc_intr"},
    {NULL, "asrc2_2risc_intr"},
    {NULL, "asrc3_2risc_intr"}, 
    {NULL, "asrc4_2risc_intr"},         
	#endif
    #endif
    #elif defined(CC_MT5398)||defined(CC_MT5880)||defined(CC_MT5860)
    // MISC2_IRQ[9]
    {NULL, "SMCARD1"},
    {NULL, "TDDC"},
    {NULL, "APWM3"},
    {NULL, "ARM9_0"},
    {NULL, "ARM9_1"},
    {NULL, "ARM9_2"},
    {NULL, "ARM9_3"},
    #elif defined(CC_MT5881)
    // MISC2_IRQ[9]
    {NULL, "SMCARD1"},
    {NULL, "TDDC"},
    {NULL, "LZMA"},
    {NULL, "NONE"},
    {NULL, "GFX3D_PP1"},
    {NULL, "GFX3D_PP1MMU"},
    {NULL, "SERDES"},
    #endif
};
#else
#error rIrqs  not defined.
#endif

#if LINUX_VERSION_CODE > KERNEL_VERSION(4, 4, 0)
extern UINT32 dtv_ko_to_virq[NR_IRQS];
extern UINT32 dtv_ko_to_hwirq[NR_IRQS];
#endif

#if (TIMEPROFILE_LOG == 1)
EXETIME_MEASURE ExeTime[40];
#endif
irqreturn_t _interrupt(int irq, void *dev_id)
{
    irqreturn_t ret;
#if LINUX_VERSION_CODE > KERNEL_VERSION(4, 4, 0)
    int hwirq = dtv_ko_to_hwirq[irq];
#else
    int hwirq = irq - SPI_OFFSET;
#endif

    if (hwirq > MAX_VECTOR_ID)
    {
        return IRQ_NONE;
    }

    if (rIrqs[hwirq].pf_isr)
    {
#if (TIMEPROFILE_LOG == 1)
    ExeTimeRecordStart(&ExeTime[hwirq].RawTimeStart);
#endif
        rIrqs[hwirq].pf_isr((UINT16)hwirq);
#if (TIMEPROFILE_LOG == 1)
    ExeTimeRecordStop(&ExeTime[hwirq] );
#endif
        ret = IRQ_HANDLED;
    }
    else
    {
        ret = IRQ_NONE;
    }

    return ret;
}

/*----------------------------------------------------------------------------
 * Function: x_os_drv_reg_isr()
 *
 * Description:
 *      this API registers an ISR with its vector id. it performs
 *      1. check is vector table is established.
 *      2. locate vector table entry from vector id.
 *      3. save the new ISR if the table entry is not occupied or is
 *         occupied by the same vector, otherwise, search forward to
 *         find one until entire table is searched.
 *
 * Inputs:
 *      ui2_vec_id: an vector id to register an ISR.
 *      pf_isr: pointer to a ISR to set.
 *      ppf_old_isr: pointer to hold the current ISR setting.
 *
 * Outputs: -
 *
 * Returns:
 *      OSR_DRV_OK : routine was successful.
 *      OSR_DRV_NOT_INIT: invalid argument.
 *      OSR_DRV_FAIL: OS driver related error.
 *---------------------------------------------------------------------------*/
INT32
x_os_drv_reg_isr(UINT16             ui2_vec_id,
                 x_os_drv_isr_fct   pf_isr,
                 x_os_drv_isr_fct   *ppf_old_isr)
{
    int result;
    INT32 ret = OSR_DRV_OK;
    unsigned long flags;
#if LINUX_VERSION_CODE > KERNEL_VERSION(4, 4, 0)
    UINT32 virq = dtv_ko_to_virq[ui2_vec_id];
#else
    UINT32 virq = ui2_vec_id+SPI_OFFSET;
#endif

    if (ui2_vec_id> MAX_VECTOR_ID)
    {
        return OSR_DRV_FAIL;
    }

    // disable IRQ.
    local_irq_save(flags);

#if 0//defined(CONFIG_SMP)
    // For SMP, also need to disable specific IRQ to prevent it run from other cores.
    // We can't use x_crit_start(), otherwise it might deadlock when:
    //  1. core 2 running reg_isr with lock held
    //  2. core 1 got IRQ, in ISR it need to enter critical section, wait for core2.
    //  3. core 2 running free_irq, which need to wait for core 1 ISR to finish.
    {
    int irqen = BIM_IsIrqEnabled(ui2_vec_id);
	if (irqen)
        BIM_DisableIrq(ui2_vec_id);
#endif

    *ppf_old_isr = rIrqs[ui2_vec_id].pf_isr;
    rIrqs[ui2_vec_id].pf_isr = pf_isr;

    if (*ppf_old_isr)
    {
        free_irq(virq, &rIrqs[ui2_vec_id].ucName);
    }

    if (pf_isr != NULL)
    {
        unsigned long irq_flags;

        irq_flags = SA_SHIRQ;

#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 10, 0)
        if ((ui2_vec_id == VECTOR_SERIAL) ||
            (ui2_vec_id == VECTOR_GFX) ||
            (ui2_vec_id == VECTOR_PDWNC))
        {
            irq_flags |= IRQF_SAMPLE_RANDOM;
        }
#endif

        result = request_irq(virq, _interrupt, irq_flags, rIrqs[ui2_vec_id].ucName, &rIrqs[ui2_vec_id].ucName);
        if (result)
        {
            ret = OSR_DRV_FAIL;
            goto out;
        }
    }

out:

#if 0//defined(CONFIG_SMP)
    if (irqen)
        BIM_EnableIrq(ui2_vec_id);
    }
#endif

    local_irq_restore(flags);

    return ret;
}

#if 1
INT32
x_os_drv_reg_isr_ex(UINT16             ui2_vec_id,
                    x_os_drv_isr_fct   pf_isr,
                    x_os_drv_isr_fct   *ppf_old_isr)
{
    int result;
#if LINUX_VERSION_CODE > KERNEL_VERSION(4, 4, 0)
    UINT32 virq = dtv_ko_to_virq[ui2_vec_id];
#else
    UINT32 virq = ui2_vec_id+SPI_OFFSET;
#endif

    if (ui2_vec_id > MAX_VECTOR_ID)
    {
        return OSR_DRV_FAIL;
    }

    *ppf_old_isr = rIrqs[ui2_vec_id].pf_isr;
    rIrqs[ui2_vec_id].pf_isr = pf_isr;

    if (*ppf_old_isr)
    {
        free_irq(virq, NULL);
    }

    if (pf_isr != NULL)
    {
        result = request_irq(virq, _interrupt, SA_INTERRUPT, rIrqs[ui2_vec_id].ucName, NULL);
        if (result)
        {
            return OSR_DRV_FAIL;
        }
    }

    return OSR_DRV_OK;
}
#endif

/*----------------------------------------------------------------------------
 * Function: os_drv_isr_init()
 *
 * Description:
 *      this API registers an ISR with its vector id. it performs
 *      1. allocate vector table
 *      2. initialize each table entry.
 *
 * Inputs:  -
 *
 * Outputs: -
 *
 * Returns:
 *      OSR_DRV_OK : routine was successful.
 *      OSR_DRV_FAIL: OS driver related error.
 *---------------------------------------------------------------------------*/
INT32
os_drv_isr_init(void)
{
    return OSR_DRV_OK;
}
